# ------------------------------------------------------------------------------
import os, os.path, sys, re, time, random, types, base64, urllib
from DateTime import DateTime
from ZODB.POSException import ConflictError
from appy import Object
import appy.gen
from appy.gen import String, Page
from appy.fields.search import Search, UiSearch, Criteria
from appy.gen.layout import ColumnLayout
from appy.gen import utils as gutils
from appy.gen.mixins import BaseMixin
from appy.gen.wrappers import AbstractWrapper
from appy.gen.descriptors import ClassDescriptor
from appy.gen.navigate import Siblings
from appy.shared import mimeTypes
from appy.shared import utils as sutils
from appy.shared.data import languages
from appy.shared.ldap_connector import LdapConnector
try:
    from AccessControl.ZopeSecurityPolicy import _noroles
except ImportError:
    _noroles = []

# Global JS internationalized messages that will be computed in every page -----
jsMessages = ('no_elem_selected', 'action_confirm', 'save_confirm',
              'warn_leave_form', 'workflow_comment')

# Error messages ---------------------------------------------------------------
PAGE_NOT_FOUND = '404: %s not found.'
PAGE_UNAUTHORIZED = 'unauthorized/problematic access to %s.'
USER_NOT_FOUND = 'user %s not found. Site change at the same URL, or ' \
  'deletion of a connected user ?'

# ------------------------------------------------------------------------------
class ToolMixin(BaseMixin):
    _appy_meta_type = 'Tool'
    xhtmlEncoding = 'text/html;charset=UTF-8'

    def setResponseHeaders(self, resp, lang):
        '''Sets response headers for any HTTP p_resp(onse) to the browser'''
        set = resp.setHeader
        set('Content-type', self.xhtmlEncoding);
        set('Content-Language', lang);
        set('Cache-Control', 'no-cache, no-store, must-revalidate');
        set('Expires', '0')

    def getPortalType(self, metaTypeOrAppyClass):
        '''Returns the name of the portal_type that is based on
           p_metaTypeOrAppyClass.'''
        appName = self.getProductConfig().PROJECTNAME
        res = metaTypeOrAppyClass
        if not isinstance(metaTypeOrAppyClass, basestring):
            res = gutils.getClassName(metaTypeOrAppyClass, appName)
        if res.find('_wrappers') != -1:
            elems = res.split('_')
            res = '%s%s' % (elems[1], elems[4])
        if res in ('User', 'Group', 'Translation'): res = appName + res
        return res

    def home(self):
        '''Returns the content of px ToolWrapper.pxHome'''
        tool = self.appy()
        return tool.pxHome({'obj': None, 'tool': tool})

    def info(self):
        '''Returns the content of px ToolWrapper.pxHome'''
        tool = self.appy()
        return tool.pxInfo({'obj': None, 'tool': tool})

    def query(self):
        '''Returns the content of px ToolWrapper.pxQuery'''
        tool = self.appy()
        return tool.pxQuery({'obj': None, 'tool': tool})

    def search(self):
        '''Returns the content of px ToolWrapper.pxSearch'''
        tool = self.appy()
        return tool.pxSearch({'obj': None, 'tool': tool})

    def getHomePage(self):
        '''Return the home page when a user hits the app'''
        # Get the currently logged user
        user = self.getUser()
        tool = self.appy()
        # Display an error if there is a problem with the authentication context
        if not user.isAnon():
            authContext = self.getProductConfig(True).authContext
            if authContext and not authContext._check(tool):
                info = self.translate('auth_context_error')
                return '%s/info?portlet=0&info=%s' % (tool.url, info)
        # If the user must change its password, redirect him to this page
        if user.changePasswordAtNextLogin:
            return '%s/edit?page=passwords' % user.url
        # If the app defines a method "getHomePage", call it
        url = None
        try:
            url = tool.getHomePage()
        except AttributeError:
            pass
        if not url:
            # Bring Managers to the config, lead others to pxHome
            if user.hasRole('Manager'):
                url = self.absolute_url()
            else:
                url = '%s/home' % self.absolute_url()
        # Does the user come to its home page for authenticating and,
        # afterwards, being redirected to some other page?
        gotoUrl = self.REQUEST.get('goto')
        if gotoUrl:
            gotoUrl = urllib.quote(gotoUrl)
            if '?' in url: url += '&goto=%s' % gotoUrl
            else: url += '?goto=%s' % gotoUrl
            self.say(self.translate('please_authenticate'))
        return url

    def getErrorPage(self, options):
        '''Gets the error page'''
        req = self.REQUEST
        login = getattr(req, 'userLogin', 'anon')
        # Add Appy-specific context
        tool = self.appy()
        options['obj'] = None
        options['tool'] = tool
        errorVal = options['error_value']
        errorType = options['error_type']
        # Manage the special case of inner errors produced by pod buffers
        if errorType == 'EvaluationError':
            errorVal = errorVal.originalError
            errorType = errorVal.__class__.__name__
        url = req['AUTHENTICATION_PATH'] or req['PATH_INFO']
        # When called from a Ajax request, avoid displaying the whole page,
        # with banner, etc: simulate a popup.
        if url.endswith('/ajax'): req['popup'] = '1'
        params = req['QUERY_STRING']
        if params: url += '?%s' % params
        # Make sure no message will be returned in the AppyMessage cookie
        if hasattr(req, 'messages'): del req.messages
        req['AppyMessage'] = None
        # Manage the various kinds of errors
        if errorType == 'NotFound':
            # Return a 404 page
            options['msg'] = self.translate('not_found')
            self.log(PAGE_NOT_FOUND % url, type='warning')
            req.RESPONSE.setStatus(404)
            return tool.px404(options).encode('utf-8')
        elif errorType == 'Unauthorized':
            if login == 'anon':
                # Let the user authenticate himself
                url = '%s?goto=%s' % (self.getSiteUrl(), urllib.quote(url))
                return self.goto(url, msg='Please authenticate')
            else:
                # Return a specific error page
                error = options['error_value']
                orig = getattr(error, 'originalError', None)
                msg = getattr(orig, 'message', '') or \
                      getattr(error, 'message', '')
                options['msg'] = msg or self.translate('unauthorized')
                self.log(PAGE_UNAUTHORIZED % url, type='warning')
                return tool.px403(options).encode('utf-8')
        else:
            # Return the standard error page
            if isinstance(errorVal, gutils.MessageException):
                # Display a nice and not a technical error message
                options['msg'] = str(errorVal)
                options['showTraceback'] = False
            else:
                # Determine what message to display. A ConflictError is an
                # unresolved ZODB (read or write) conflict error.
                if isinstance(errorVal, ConflictError):
                    label = 'conflict_error'
                    msg = str(errorVal) # Do not dump the entire traceback
                else:
                    label = 'server_error'
                    msg = sutils.Traceback.get()
                options['msg'] = self.translate(label)
                self.log(msg, type='error')
                # Integrate the traceback if the user is admin
                options['showTraceback'] = login == 'admin'
            return tool.px500(options).encode('utf-8')

    def getHomeObject(self, inPopup=False):
        '''The concept of "home object" is the object where the user must "be",
           even if he is "nowhere". For example, if the user is on a search
           screen, there is no contextual object. In this case, if we have a
           home object for him, we will use it as contextual object, and its
           portlet menu will nevertheless appear: the user will not have the
           feeling of being lost.'''
        # If we are in the popup, we do not want any home object in the way.
        if inPopup: return
        # If the app defines a method "getHomeObject", call it.
        try:
            return self.appy().getHomeObject()
        except AttributeError:
            # For managers, the home object is the config. For others, there is
            # no default home object.
            user = self.getUser()
            if user.hasRole('Manager'): return self.appy()

    def getPageTitle(self, obj):
        '''Returns content for tag html->head->title'''
        # The page title is based on the p_obj title
        res = (obj and (obj != self)) and obj.getShownValue() or \
              self.translate('app_name')
        # Ensure this is a string
        if not isinstance(res, basestring): res = str(res)
        # Remove <br/> tags
        if '<br/>' in res: res = res[:res.index('<br/>')]
        # Add the page title if found
        if obj and ('page' in self.REQUEST):
            pageLabel = '%s_page_%s' % (obj.meta_type, self.REQUEST['page'])
            text = self.translate(pageLabel, blankOnError=True)
            if text:
                if res: res += ' :: '
                res += text
        try:
            res = self.truncateValue(res, width=100)
        except UnicodeDecodeError:
            res = '?'
        return res

    def getCatalog(self):
        '''Returns the catalog object'''
        return self.getParentNode().catalog

    def getCatalogValue(self, obj, indexName):
        '''Get, for p_obj, the value stored in the catalog for the index
           named p_indexName.'''
        catalogBrain = self.getObject(obj.id, brain=True)
        catalog = self.getApp().catalog
        index = catalog.Indexes[indexName]
        indexType = index.getTagName()
        if indexType == 'ZCTextIndex':
            # Zope bug: the lexicon can't be retrieved correctly
            index._v_lexicon = getattr(catalog, index.lexicon_id)
        res = index.getEntryForObject(catalogBrain.getRID())
        if indexType == 'DateIndex':
            # The index value is a number. Add a DateTime representation too.
            if not res: res = ''
            else:
                from appy.fields.date import getDateFromIndexValue
                res = '%d (%s)' % (res, getDateFromIndexValue(res))
        return res

    def getApp(self):
        '''Returns the root Zope object'''
        return self.getPhysicalRoot()

    def getSiteUrl(self):
        '''Returns the absolute URL of this site'''
        return self.getApp().absolute_url()

    def getIncludeUrl(self, name, bg=False):
        '''Gets the full URL of an external resource, like an image, a
           Javascript or a CSS file, named p_name. If p_bg is True, p_name is
           an image that is meant to be used in a "style" attribute for defining
           the background image of some XHTML tag.'''
        # If no extension is found in p_name, we suppose it is a png image
        if '.' not in name: name += '.png'
        url = '%s/ui/%s' % (self.getPhysicalRoot().absolute_url(), name)
        if not bg: return url
        return 'background-image: url(%s)' % url

    def doPod(self):
        '''Performs an action linked to a pod field: generate, freeze,
           unfreeze... a document from a pod field.'''
        rq = self.REQUEST
        # Get the object that is the target of this action. On heavy load,
        # request keys "objectUid" and "fieldName" may not be present.
        objectId = rq.get('objectUid')
        if not objectId:
            self.log('Wrong doPod request (no object ID)', type='error')
            return
        obj = self.getObject(objectId, appy=True)
        fieldName = rq.get('fieldName')
        if not fieldName:
            self.log('Wrong doPod request (no field name)', type='error')
            return
        return obj.getField(fieldName).onUiRequest(obj, rq)

    def getAppName(self):
        '''Returns the name of the application'''
        return self.getProductConfig().PROJECTNAME

    def getPath(self, path):
        '''Returns the folder or object whose absolute path p_path.'''
        res = self.getPhysicalRoot()
        if path == '/': return res
        path = path[1:]
        if '/' not in path: return res._getOb(path) # For performance
        for elem in path.split('/'): res = res._getOb(elem)
        return res

    def showGlobalSelector(self):
        '''On some pages, global selectors (like the language or authentication
           context selectors) must not be visible because they would reload the
           entire page.'''
        page = self.REQUEST.get('ACTUAL_URL').split('/')[-1]
        return page not in ('edit', 'search', 'do')

    def showLanguageSelector(self):
        '''We must show the language selector if the app config requires it and
           it there is more than 2 supported languages. Moreover, on some pages,
           switching the language is not allowed.'''
        cfg = self.getProductConfig(True)
        if not cfg.languageSelector or cfg.forcedLanguage: return
        if len(cfg.languages) < 2: return
        return self.showGlobalSelector()

    def showForgotPassword(self):
        '''We must show link "forgot password?" when the app requires it.'''
        return self.getProductConfig(True).activateForgotPassword

    def getLanguages(self):
        '''Returns the supported languages. First one is the default.'''
        return self.getProductConfig(True).languages

    def getLanguageName(self, code, lowerize=False):
        '''Gets the language name (in this language) from a 2-chars language
           p_code.'''
        res = languages.get(code)[2]
        if not lowerize: return res
        return res.lower()

    def switchLanguage(self):
        '''Sets the language cookie with the new desired language code that is
           in request["language"].'''
        rq = self.REQUEST
        rq.RESPONSE.setCookie('_ZopeLg', rq['language'], path='/')
        return self.goto(self.getReferer())

    def switchContext(self):
        '''Switch to a new context'''
        # Check that context switching is allowed
        authContext = self.getProductConfig(True).authContext
        rq = self.REQUEST
        atool = self.appy()
        ctx = rq.get('ctx', '')
        if not authContext or not authContext._maySwitchTo(atool, ctx):
            self.raiseUnauthorized()
        # Perform the switch by updating the authentication cookie
        gutils.updateCookie(rq, ctx)
        authContext._setOnRequest(atool, rq, ctx)
        # Return to the URL specified in the request, or, by default:
        # a) return to the home page if we were displaying an error message or
        #    if required by the auhentication context;
        # b) else, stay on the same page (=go to the referer page).
        back = rq.get('url')
        if not back:
            referer = self.getReferer()
            if authContext.homeAfterSwitch or ('/config/info' in referer):
                back = self.getHomePage()
            else:
                back = referer
        return self.goto(back)

    def flipLanguageDirection(self, align, dir):
        '''According to language direction p_dir ('ltr' or 'rtl'), this method
           turns p_align from 'left' to 'right' (or the inverse) when
           required.'''
        if dir == 'ltr': return align
        if align == 'left': return 'right'
        if align == 'right': return 'left'
        return align

    def getGlobalCssJs(self, dir):
        '''Returns the list of CSS and JS files to include in the main template.
           The method ensures that appy.css and appy.js come first. If p_dir
           (=language *dir*rection) is "rtl" (=right-to-left), the stylesheet
           for rtl languages is also included.'''
        names = self.getPhysicalRoot().ui.objectIds('File')
        # The single Appy Javascript file
        names.remove('appy.js'); names.insert(0, 'appy.js')
        # CSS changes for left-to-right languages
        names.remove('appyrtl.css')
        if dir == 'rtl': names.insert(0, 'appyrtl.css')
        names.remove('appy.css'); names.insert(0, 'appy.css')
        return names

    def consumeMessages(self, unlessRedirect=False):
        '''Returns the list of messages to show to a web page'''
        # Do not consume anything if p_unlessRedirect is True and we are
        # redirecting the user.
        rq = self.REQUEST
        resp = rq.RESPONSE
        if unlessRedirect and ('appy-redirect' in resp.headers): return
        # Try to get messages from the 'AppyMessage' cookie
        message = rq.get('AppyMessage', None)
        if message:
            # Dismiss the cookie
            resp.expireCookie('AppyMessage', path='/')
            return urllib.unquote(message)
        # Try to get it from rq.messages
        messages = getattr(rq, 'messages', None)
        if messages:
            del rq.messages
            return ' '.join(messages)
        return ''

    def getRootClasses(self):
        '''Returns the list of root classes for this application'''
        cfg = self.getProductConfig().appConfig
        rootClasses = cfg.rootClasses
        if rootClasses == None: return [] # No root class at all
        if not rootClasses:
            # We consider every class as being a root class
            rootClasses = self.getProductConfig().appClassNames
        return [self.getAppyClass(k) for k in rootClasses]

    def getRootFolder(self, klass):
        '''Gets the name of the root folder where to store instances of
           p_klass'''
        rootFolders = self.getProductConfig().appConfig.rootFolders
        return rootFolders.get(klass.__name__, 'data')

    def getRootFolders(self):
        '''Get names of all root folders'''
        rootFolders = self.getProductConfig().appConfig.rootFolders
        res = ['config', 'data']
        for name in rootFolders.itervalues():
            if name not in res:
                res.append(name)
        return res

    def getSearchInfo(self, className, refInfo=None):
        '''Returns, as an object:
           - the list of searchable fields (some among all indexed fields);
           - the number of columns for layouting those fields.'''
        fields = []
        if refInfo:
            # The search is triggered from a Ref field
            refObject, refField = self.getRefInfo(refInfo, nameOnly=False)
            fieldNames = refField.queryFields or ()
            nbOfColumns = refField.queryNbCols
        else:
            # The search is triggered from an app-wide search
            klass = self.getAppyClass(className)
            fieldNames = getattr(klass, 'searchFields', None)
            if callable(fieldNames): fieldNames = fieldNames(self.appy())
            nbOfColumns = getattr(klass, 'numberOfSearchColumns', 3)
        # If fields are not explicitly listed, take al indexed fields on this
        # class.
        if not fieldNames:
            fieldNames = [f.name for f in self.getAllAppyTypes(className) \
                          if f.indexed]
            # Put SearchableText in front of the list and ensure "title"
            # is not there
            if 'title' in fieldNames: fieldNames.remove('title')
            st = 'SearchableText'
            if st in fieldNames: fieldNames.remove(st)
            fieldNames.insert(0, st)
        for name in fieldNames:
            field = self.getAppyType(name, className=className)
            fields.append(field)
        return Object(fields=fields, nbOfColumns=nbOfColumns)

    def showPortlet(self, obj, layoutType):
        '''When must the portlet be shown? p_obj and p_layoutType can be None
           if we are not browsing any objet (ie, we are on the home page).'''
        # Not on 'edit' pages or if there is no root class
        classes = self.getProductConfig(True).rootClasses
        req = self.REQUEST
        if not classes or (layoutType == 'edit') or \
           (req.get('portlet', '1') == '0'): return
        res = True
        if obj and hasattr(obj, 'showPortlet'):
            res = obj.showPortlet()
        else:
            tool = self.appy()
            if hasattr(tool, 'showPortletAt'):
                res = tool.showPortletAt(req['ACTUAL_URL'])
        return res

    def showSidebar(self, obj, layoutType):
        '''The sidebar must be shown when p_obj is shown on the "view"
           p_layoutType and if obj's class declares to use the sidebar.
           If the sidebar must be shown, its width is returned.'''
        if not obj: return
        sidebar = getattr(obj.klass, 'sidebar', None)
        if not sidebar: return
        if callable(sidebar): sidebar = sidebar(self.appy())
        if not sidebar: return
        if (sidebar.show == True) or (sidebar.show == layoutType):
            return getattr(sidebar, 'width', '170px')

    def getCollapsible(self, zone, align):
        '''Gets a Collapsible instance for showing/hiding some p_zone
           ("porlet" or "sidebar").'''
        icons = (align == 'left') and 'showHide' or 'showHideInv'
        return gutils.Collapsible('appy%s' % zone.capitalize(), self.REQUEST,
                                  default='expanded', icons=icons, align=align)

    def getObject(self, uid, appy=False, brain=False, temp=False):
        '''Allows to retrieve an object from its p_uid. p_brain and p_temp
           cannot be both True'''
        app = self.getPhysicalRoot()
        if not temp:
            # Search first among wrappers: it is more performant
            req = getattr(self, 'REQUEST', None)
            if not brain and req and hasattr(req, 'wrappers') and \
               (uid in req.wrappers):
                res = req.wrappers[uid]
                if not appy:
                    res = res.o
            else:
                res = app.catalog(UID=uid)
                if not res: return
                res = res[0]
                if brain: return res
                res = res._unrestrictedGetObject()
        else:
            res = getattr(app.temp_folder, uid)
        if not appy: return res
        return res.appy()

    def toObject(self, brain, noSecurity=False, appy=False):
        '''Get an object from a catalog's p_brain. If appy is True, it returns
           an Appy object instead of a Zope object. If p_noSecurity is True,
           it gets the object even if the logged user does not have the right to
           get it.

           If no object is found for this brain, the brain is removed from the
           catalog, the operation is logged and r_ is None.'''
        # Choose the catalog method to use for getting an object from its brain
        method = noSecurity and '_unrestrictedGetObject' or 'getObject'
        try:
            r = getattr(brain, method)()
            if appy: r = r.appy()
            return r
        except KeyError:
            # Do not generate an error: ignore the dead brain. If the caller is
            # a SomeObjects instance, this entry will be removed from the
            # catalog afterwards.
            pass

    def getAllowedValue(self, user=None):
        '''Gets, for the current user (or this p_user if not None), the value of
           index "Allowed".'''
        user = user or self.getUser()
        # Get the user roles. If we do not make a copy of the list here, we will
        # really add user logins among user roles!
        res = user.getRoles()[:]
        # Get the user logins
        if user.login != 'anon':
            for login in user.getLogins():
                res.append('user:%s' % login)
        return res

    def getFilters(self, class_, filters=None):
        '''Extracts, from the request, filters defined on fields belonging to
           this p_class_'''
        r = sutils.getDictFrom(filters or self.REQUEST.get('filters'))
        # Apply a potential transform on every filter value
        for name in r.keys():
            # Get the corresponding field on p_class_
            field = getattr(class_, name, None)
            if field:
                try:
                    r[name] = field.getStorableValue(None, r[name])
                except Exception, e:
                    # The encoded value is invalid. Ignore it.
                    del r[name]
        return r

    def executeQuery(self, className, searchName=None, startNumber=0,
                     search=None, remember=False, brainsOnly=False,
                     maxResults=None, noSecurity=False, sortBy=None,
                     sortOrder='asc', filters=None, refObject=None,
                     refField=None, search2=None):
        '''Executes a query on instances of a given p_className in the catalog.
           If p_searchName is specified, it corresponds to:
             1) a search defined on p_className: additional search criteria
                will be added to the query, or;
             2) "customSearch": in this case, additional search criteria will
                also be added to the query, but those criteria come from the
                session (in key "searchCriteria") and came from pxSearch.

           More criteria can be given via an additional p_search2.

           Objects are retrieved from p_startNumber. If p_search is defined,
           it corresponds to a custom Search instance (instead of a predefined
           named search like in p_searchName). If both p_searchName and p_search
           are given, p_search is ignored.

           This method returns a list of objects in the form of an instance of
           SomeObjects (see in appy.gen.utils). If p_brainsOnly is True, it
           returns a list of brains instead (can be useful for some usages like
           knowing the number of objects without needing to get information
           about them). If no p_maxResults is specified, the method returns
           maximum self.numberOfResultsPerPage. The method returns all objects
           if p_maxResults equals string "NO_LIMIT".

           If p_noSecurity is True, it gets all the objects, even those that the
           currently logged user can't see.

           The result is sorted according to the potential sort key defined in
           the Search instance (Search.sortBy, together with Search.sortOrder).
           But if parameter p_sortBy is given, it defines or overrides the sort.
           In this case, p_sortOrder gives the order (*asc*ending or
           *desc*ending).

           If p_filters are given (as a dict), it represents additional search
           parameters to take into account.

           If p_refObject and p_refField are given, the query is limited to the
           objects that are referenced from p_refObject through p_refField.'''
        params = {'ClassName': className}
        klass = self.getAppyClass(className, wrapper=True)
        if not brainsOnly: params['batch'] = True
        # Manage additional criteria from a search when relevant
        if searchName: search = self.getSearch(className, searchName)
        if search:
            # Add in params search and sort criteria
            search.updateSearchCriteria(self, params, klass)
        if search2: search2.updateSearchCriteria(self, params, klass)
        # Determine or override sort if specified
        if sortBy:
            params['sort_on'] = Search.getIndexName(sortBy, klass, usage='sort')
            params['sort_order'] = (sortOrder == 'desc') and 'reverse' or None
        # If defined, add the filter among search parameters
        if filters:
            for key, value in filters.iteritems():
                key = Search.getIndexName(key, klass, usage='filter')
                value = Search.getSearchValue(key, value, klass)
                params[key] = value
        # Limit the query to some ref when relevant
        if refObject:
            refField = refObject.getAppyType(refField)
            back = refField.back
            if back: params['Container'] = '%s_%s' % (refObject.id, back.name)
        # Use index "Allowed" if noSecurity is False
        if not noSecurity and ('Allowed' not in params):
            params['Allowed'] = self.getAllowedValue()
        # Perform the query
        brains = self.getPath('/catalog')(**params)
        # Compute maxResults
        if not maxResults:
            if refField: maxResults = refField.maxPerPage
            else: maxResults = search.maxPerPage or \
                               self.appy().numberOfResultsPerPage
        elif maxResults == 'NO_LIMIT':
            maxResults = None
        # Return brains only if required
        if brainsOnly:
            if not maxResults: return brains
            else: return brains[:maxResults]
        res = gutils.SomeObjects(brains, maxResults, startNumber,
                                 noSecurity=noSecurity)
        res.brainsToObjects(self)
        # In some cases (p_remember=True), we need to keep some information
        # about the query results in the current user's session, allowing him
        # to navigate within elements without re-triggering the query every
        # time a page for an element is consulted.
        if remember:
            if not searchName:
                if not search:
                    searchName = className
                else:
                    searchName = search.getSessionKey(className, full=False)
            uids = {}
            i = -1
            for obj in res.objects:
                i += 1
                uids[startNumber+i] = obj.id
            self.REQUEST.SESSION['search_%s' % searchName] = uids
        return res

    def getResultCss(self, className, layoutType='view'):
        '''Returns the CSS class(es) to apply when displaying a table containing
           instances of class named p_className.'''
        k = self.getAppyClass(className)
        if hasattr(k, 'listCss'): return k.listCss
        if layoutType == 'cell': return 'list'
        return 'list compact'

    def truncateValue(self, value, width=20, suffix=u'...'):
        '''Truncates the p_value according to p_width. p_value has to be
           unicode-encoded for being truncated (else, one char may be spread on
           2 chars).'''
        # Param p_width can be None
        if not width: width = 20
        if isinstance(value, str): value = value.decode('utf-8')
        if len(value) > width:
            value = value[:width] + suffix
        return value.encode('utf-8')

    def truncateText(self, text, width=20):
        '''Truncates p_text to max p_width chars. If the text is longer than
           p_width, the truncated part is put in a "abbr" html tag. p_text
           has to be unicode-encoded for being truncated (else, one char may be
           spread on 2 chars).'''
        # Param p_width can be None
        if not width: width = 20
        if isinstance(text, str): text = text.decode('utf-8')
        if len(text) > width:
            text = u'<abbr title="%s">%s...</abbr>' % (text, text[:width])
        return text.encode('utf-8')

    def splitList(self, l, sub):
        '''Returns a list made of the same elements as p_l, but grouped into
           sub-lists of p_sub elements.'''
        return sutils.splitList(l, sub)

    def quote(self, s, escapeWithEntity=True):
        '''Returns the quoted version of p_s.'''
        if not isinstance(s, basestring): s = str(s)
        repl = escapeWithEntity and '&apos;' or "\\'"
        s = s.replace('\r\n', '').replace('\n', '').replace("'", repl)
        return "'%s'" % s

    def getLayoutType(self, req):
        '''Guess the current layout type, according to actual URL and
           parameters.'''
        r = self.REQUEST.get('layoutType')
        if r: return r
        url = self.REQUEST['ACTUAL_URL']
        if url.endswith('/view'):
            r = 'view'
        elif url.endswith('/edit'):
            r = 'edit'
        elif url.endswith('/do'):
            r = ('search' in req) and 'search' or 'edit'
        else:
            r = None
        return r

    def getZopeClass(self, name):
        '''Returns the Zope class whose name is p_name.'''
        exec 'from Products.%s.%s import %s as C'% (self.getAppName(),name,name)
        return C

    def getAppyClass(self, zopeName, wrapper=False):
        '''Gets the Appy class corresponding to the Zope class named p_name.
           If p_wrapper is True, it returns the Appy wrapper. Else, it returns
           the user-defined class.'''
        # p_zopeName may be the name of the Zope class *or* the name of the Appy
        # class (shorter, not prefixed with the underscored package path).
        classes = self.getProductConfig().allShortClassNames
        if zopeName in classes: zopeName = classes[zopeName]
        zopeClass = self.getZopeClass(zopeName)
        if wrapper: return zopeClass.wrapperClass
        else: return zopeClass.wrapperClass.__bases__[-1]

    def getSearchAdvanced(self, klass):
        '''Gets the Search instance representing default search parameters for
           p_klass' searches (advanced, all and live searches).'''
        search = getattr(klass, 'searchAdvanced', None)
        if not search: return
        if callable(search): search = search(self.appy())
        return search

    def getAllClassNames(self):
        '''Returns the name of all classes within this app, including default
           Appy classes (Tool, Translation, Page, etc).'''
        return self.getProductConfig().allClassNames + [self.__class__.__name__]

    def getCreateFor(self, klass):
        '''How to create instances of p_klass ? The answer can be: via a
           web form ("form"), via a form pre-filled from a template (a Search
           instance) or programmatically only (None).'''
        # By default, instances are created via an empty web form
        res = getattr(klass, 'create', 'form')
        if callable(res): res = res(self.appy())
        return res

    def getCreateLink(self, className, create, formName, sourceField=None,
                      insert=None):
        '''When instances of p_className must be created from a template, this
           method returns the link allowing to query such templates from a
           popup.'''
        res = '%s/query?className=%s&search=fromSearch&popup=1&fromClass=%s' \
               '&formName=%s' % (self.appy().url,
               self.getPortalType(create.klass), className, formName)
        # When object creation occurs via a Ref field, its coordinates are given
        # in p_sourceField as a string: "sourceObjectId:fieldName".
        if sourceField: res += '&sourceField=%s' % sourceField
        # p_insert is filled when the object must be inserted at a given place
        if insert: res += '&insert=%s' % insert
        return res

    def userMaySearch(self, klass, layoutType):
        '''May the user search among instances of root p_klass ?'''
        # When editing a form, one should avoid annoying the user with this
        if layoutType == 'edit': return
        if hasattr(klass, 'maySearch'): return klass.maySearch(self.appy())
        return True

    def userMayCreate(self, klass, checkInitiator=False, initiator=None):
        '''May the logged user create instances of p_klass ? This information
           can be defined on p_klass, in static attribute "creators".
           1. If this attr holds a list, we consider it to be a list of roles,
              and we check that the user has at least one of those roles.
           2. If this attr holds a boolean, we consider that the user can create
              instances of this class if the boolean is True.
           3. If this attr stores a method, we execute the method, and via its
              result, we fall again in cases 1 or 2.

           If p_klass does not define this attr "creators", we will use a
           default list of roles as defined in the config.'''
        # An initiator object may dictate his own rules for instance creation
        if checkInitiator:
            init = initiator or self.getInitiator()
            if init and getattr(init.field, 'creators', None):
                return self.getUser().hasRole(init.field.creators, init.obj)
        # Get the value of attr "creators", or a default value if not present
        if hasattr(klass, 'creators'):
            creators = klass.creators
        else:
            creators = self.getProductConfig().appConfig.defaultCreators
        # Resolve case (3): if "creators" is a method, execute it.
        if callable(creators): creators = creators(self.appy())
        # Resolve case (2)
        if isinstance(creators, bool) or not creators: return creators
        # Resolve case (1): checks whether the user has at least one of the
        # roles listed in "creators".
        for role in self.getUser().getRoles():
            if role in creators:
                return True

    def subTitleIsUsed(self, className):
        '''Does class named p_className define a method "getSubTitle" ?'''
        klass = self.getAppyClass(className)
        return hasattr(klass, 'getSubTitle')

    def onSearchObjects(self):
        '''Called when the user triggers a search from pxSearch'''
        # Retrieve search criteria from the request
        req = self.REQUEST
        criteria = Criteria(self)
        criteria.getFromRequest()
        # Renders the PX displaying search results
        req['search'] = 'customSearch'
        req['criteria'] = criteria.asString()
        # The "string" criteria must be present in the request for performing
        # subsequent ajax searches. But for the current response, criteria as a
        # dict are already there. Because we cache it as soon as we evaluate it
        # from their "string" version, here we cheat and already set the cache
        # on the request with the dict. So we avoid evaluating the "string"
        # version.
        req.searchCriteria = criteria.criteria
        return self.query()

    def getJavascriptMessages(self, cfg):
        '''Returns the translated version of messages that must be shown in
           Javascript popups.'''
        r = ['var wrongTextInput="%s none";' % cfg.ui.wrongTextColor]
        for msg in jsMessages:
            r.append('var %s="%s";' % (msg, self.translate(msg)))
        return '\n'.join(r)

    # Standard Appy column specifiers
    standardColumnSpecifiers = Object(
      number = Object(special=True, field='_number', width='15px',
                      align='left', header=False),
      checkboxes = Object(special=True, field='_checkboxes', width='10px',
                          align='center', header=True)
    )
    def getColumnsSpecifiers(self, className, columnLayouts, dir,
                             addNumber=False, addCheckboxes=False):
        '''Extracts and returns, from a list of p_columnLayouts, info required
           for displaying columns of field values for instances of p_className,
           either in a result screen or for a Ref field.
        '''
        # p_columnLayouts are specified for each field whose values must be
        # shown. 2 more, not-field-related, column layouts can be specified with
        # these names:
        # ----------------------------------------------------------------------
        # "_number"     | if the listed objects must be numbered by Appy, this
        #               | string represents the column containing that number;
        # ----------------------------------------------------------------------
        # "_checkboxes" | if Appy must show checkboxes for the listed objects,
        #               | this string represents the column containing the
        #               | checkboxes.
        # ----------------------------------------------------------------------
        # If columns "_number" and "_checkboxes" are not among p_columnLayouts
        # but are required (by p_addNumber and p_addCheckboxes), they are added
        # to the result. Specifying them within p_columnLayouts allows to give
        # them a precise position among all columns. When automatically added,
        # they will appear before any other column (which is desirable in most
        # cases).
        r = []
        numberFound = checkboxesFound = False
        for info in columnLayouts:
            name, width, align, header = ColumnLayout(info).get()
            # It that a special column name ?
            special = True
            if name == '_number': numberFound = True
            elif name == '_checkboxes': checkboxesFound = True
            else: special = False
            align = self.flipLanguageDirection(align, dir)
            # For non-special columns, get the corresponding field
            if not special:
                field = self.getAppyType(name, className)
                if not field:
                    self.log('field "%s", used in a column specifier, not ' \
                             'found.' % name, type='warning')
                    continue
            else:
                # Let the column name in attribute "field"
                field = name
            r.append(Object(special=special, field=field, width=width,
                            align=align, header=header))
        # Add special columns if required and not present
        if addCheckboxes and not checkboxesFound:
            r.insert(0, self.standardColumnSpecifiers.checkboxes)
        if addNumber and not numberFound:
            r.insert(0, self.standardColumnSpecifiers.number)
        return r

    def getRefInfo(self, info=None, nameOnly=True):
        '''When a search is restricted to objects referenced through a Ref
           field, this method returns information about this reference: the
           source object and the Ref field.'''
        if not info:
            criteria = Criteria.readFromRequest(self)
            if criteria and ('_ref' in criteria): info = criteria['_ref']
        if not info: return None, None
        id, field = info.split(':')
        obj = self.getObject(id)
        if not nameOnly:
            field = obj.getAppyType(field)
        return obj, field

    def getDynamicSearches(self, klass, req):
        '''Gets the dynamic searches defined on p_klass, or get them cached on
           the p_req(uest) object.'''
        # Create the cache if it does not exist yet
        if not hasattr(req, 'dynamicSearches'):
            req.dynamicSearches = dyn = {}
        else:
            dyn = req.dynamicSearches
        # Get the cache version for p_klass if it exists, or really compute
        # dynamic searches and add it in the cache.
        key = klass.__name__
        if key in dyn:
            return dyn[key]
        else:
            # Get it or create an empty entry if the class does not define it
            if hasattr(klass, 'getDynamicSearches'):
                dyn[key] = r = klass.getDynamicSearches(self.appy())
                return r
            else:
                dyn[key] = None

    def getGroupedSearches(self, klass):
        '''Returns an object with 2 attributes:
           * "searches" stores the searches that are defined for p_klass;
           * "default" stores the search defined as the default one.
           Every item representing a search is a dict containing info about a
           search or about a group of searches.
        '''
        res = []
        default = None # Also retrieve the default one here
        groups = {} # The already encountered groups
        page = Page('searches') # A dummy page required by class UiGroup
        # Get the searches statically defined on the class
        className = self.getPortalType(klass)
        searches = ClassDescriptor.getSearches(klass, tool=self.appy())
        # Get the dynamically computed searches
        dyn = self.getDynamicSearches(klass, self.REQUEST)
        if dyn: searches += dyn
        for search in searches:
            # Create the search descriptor
            uiSearch = UiSearch(search, className, self)
            if not search.group:
                # Insert the search at the highest level, not in any group
                res.append(uiSearch)
            else:
                uiGroup = search.group.insertInto(res, groups, page, className,
                                                  content='searches')
                uiGroup.addElement(uiSearch)
            # Is this search the default search?
            if search.default: default = uiSearch
        return Object(searches=res, default=default)

    def getSearch(self, className, name, ui=False):
        '''Gets the Search instance (or a UiSearch instance if p_ui is True)
           corresponding to the search named p_name, on class p_className.'''
        initiator = None
        req = self.REQUEST
        if name == 'customSearch':
            # It is a custom search whose parameters are marshalled in key
            # "criteria". Unmarshal them.
            criteria = Criteria.readFromRequest(self) or {}
            res = Search('customSearch', **criteria)
            # Avoid name clash
            if 'group' in criteria: res.fields['group'] = criteria['group']
            # Take into account default params for the advanced search
            advanced = self.getSearchAdvanced(self.getAppyClass(className))
            if advanced: res.merge(advanced)
        elif (name == 'allSearch') or not name:
            # It is the search for every instance of p_className
            res = Search('allSearch')
            name = res.name
            # Take into account default params for the advanced search
            advanced = self.getSearchAdvanced(self.getAppyClass(className))
            if advanced: res.merge(advanced)
        elif name == 'fromSearch':
            # It is the search for selecting a template for creating an object
            fromClass = req.get('fromClass')
            sourceField = req.get('sourceField')
            if sourceField:
                # We are creating an object from a Ref field
                objectId, fieldName = sourceField.split(':')
                obj = self.getObject(objectId)
                res = obj.getAppyType(fieldName).getAttribute(obj, 'create')
            else:
                # We are creating a root object
                res = self.getCreateFor(self.getAppyClass(fromClass))
            initiator = UiSearch.TemplateInitiator(fromClass,
              req.get('formName'), req.get('insert'), sourceField)
        elif ',' in name:
            # The search is defined in a field
            id, fieldName, mode = name.split(',')
            # Get the object with this "id". In the case of a Ref field with
            # link=popup, the initiator object can be a temp object.
            obj = self.getObject(id, appy=True) or \
                  self.getObject(id, appy=True, temp=True)
            field = obj.getField(fieldName)
            if field.type == 'Ref':
                initiator = UiSearch.RefInitiator(obj, field, fieldName, mode)
                res = field.getSelect(obj)
            elif field.type == 'Computed':
                res = field.getSearch(obj)
        else:
            appyClass = self.getAppyClass(className)
            # Search among static searches
            res = ClassDescriptor.getSearch(appyClass, name)
            if not res:
                # Search among dynamic searches
                dyn = self.getDynamicSearches(appyClass, req)
                if dyn:
                    for search in dyn:
                        if search.name == name:
                            res = search
                            break
        # The search may not exist
        if not res: self.raiseUnauthorized(self.translate('search_broken'))
        # Return a UiSearch if required
        if ui: res = UiSearch(res, className, self, initiator, name)
        return res

    def getPxTarget(self, zobj, px):
        '''When a p_px is requested from Ajax, this method gets the target PX
           field or search when relevant.'''
        res = Object(className=None, name=None, pxName=None, field=None)
        if len(px) == 3:
            res.className, res.name, res.pxName = px
        elif len(px) == 2:
            res.name, res.pxName = px
        else:
            # Only the PX name: the PX is not field or search-related
            res.pxName = px
            return res
        # Find the field or search named "res.name" and put it in "res.field"
        if not res.className:
            # If no class name is specified, it can only be a field
            res.field = zobj.getAppyType(res.name)
        else:
            # Else, it can be a field or a search
            res.field = zobj.getAppyType(res.name, res.className) or \
                        self.getSearch(res.className, res.name, ui=True)
        return res

    def getLiveSearch(self, klass, keywords):
        '''Gets the Search instance for performing a live search on p_klass'''
        if not keywords.endswith('*'): keywords += '*'
        res = Search('liveSearch', SearchableText=keywords)
        # Take into account default params for the advanced search
        advanced = self.getSearchAdvanced(klass)
        if advanced: res.merge(advanced)
        return res

    def advancedSearchEnabledFor(self, klass):
        '''Is advanced search visible for p_klass ?'''
        search = self.getSearchAdvanced(klass)
        # By default, advanced search is enabled
        if not search: return True
        # Evaluate attribute "show" on this Search instance representing the
        # advanced search.
        return search.isShowable(klass, self.appy())

    def portletBottom(self, klass):
        '''Is there a custom zone to display at the bottom of the portlet zone
           for p_klass?'''
        if not hasattr(klass, 'getPortletBottom'): return ''
        res = klass.getPortletBottom(self.appy())
        if not res: return ''
        return res

    def getNavigationInfo(self, nav, inPopup):
        '''Produces a Siblings instance from navigation info p_nav'''
        return Siblings.get(nav, self, inPopup)

    def getGroupedSearchFields(self, searchInfo):
        '''This method transforms p_searchInfo.fields, which is a "flat"
           list of fields, into a list of lists, where every sub-list having
           length p_searchInfo.nbOfColumns. For every field, scolspan
           (=colspan "for search") is taken into account.'''
        res = []
        row = []
        rowLength = 0
        for field in searchInfo.fields:
            # Can I insert this field in the current row?
            remaining = searchInfo.nbOfColumns - rowLength
            if field.scolspan <= remaining:
                # Yes.
                row.append(field)
                rowLength += field.scolspan
            else:
                # We must put the field on a new line. Complete the current one
                # if not complete.
                while rowLength < searchInfo.nbOfColumns:
                    row.append(None)
                    rowLength += 1
                res.append(row)
                row = [field]
                rowLength = field.scolspan
        # Complete the last unfinished line if required
        if row:
            while rowLength < searchInfo.nbOfColumns:
                row.append(None)
                rowLength += 1
            res.append(row)
        return res

    # --------------------------------------------------------------------------
    # Authentication-related methods
    # --------------------------------------------------------------------------
    def _getLogin(self, req):
        '''Get the login as encoded in the UI by the user and apply tranform(s)
           to it when relevant.'''
        r = req.get('__ac_name')
        if r:
            r = r.strip()
            transform = self.getProductConfig(True).loginTransform
            if transform:
                exec 'r = r.%s()' % transform
        return r

    def _disableUser(self, req):
        '''Disable the authentication cookie and remove credentials stored on
           the p_req(uest).'''
        req.RESPONSE.expireCookie('_appy_', path='/')
        k = 'HTTP_AUTHORIZATION'
        req._auth = req[k] = req._orig_env[k] = None
        

    def identifyUser(self, alsoSpecial=False):
        '''To identify a user means: get its login, password and authentication
           context. There are several places to look for this information: http
           authentication, cookie of credentials coming from the web form.

           If no user could be identified, and p_alsoSpecial is True, we will
           nevertheless identify a "special user": "system", representing the
           system itself (running at startup or in batch mode) or "anon",
           representing an anonymous user.'''
        tool = self.appy()
        req = tool.request
        login = password = ctx = place = None
        # a. Identify the user from HTTP basic authentication
        if getattr(req, '_auth', None):
            # Credentials are present (used by a peer application, or the ZMI).
            # Decode it.
            creds = req._auth
            if creds.lower().startswith('basic '):
                try:
                    creds = creds.split(' ')[-1]
                    login, password = base64.decodestring(creds).split(':', 1)
                    if login: place = 'basic'
                except Exception, e:
                    pass
        # b. Identify the user from the authentication cookie
        if not login:
            try:
                login, password, ctx = gutils.readCookie(req)
            except Exception, e:
                self.log('Unreadable cookie (%s)' % str(e), type='error',
                         noUser=True)
            if login: place = 'cookie'
        # c. Identify the user from the authentication form
        if not login:
            login = self._getLogin(req)
            password = req.get('__ac_password', '') or ''
            ctx = req.get('__ac_ctx')
            if login: place = 'form'
        # d. Identify the user from a SSO reverse proxy
        if not login or (login == '_s_s_o_'):
            sso = self.getProductConfig(True).sso
            if sso:
                login = sso.extractUserLogin(tool, req, warn=not alsoSpecial)
                if login: place = 'sso'
        # Stop identification here if we don't need to return a special user
        if not alsoSpecial: return login, password, ctx, place
        # d. All the identification methods failed. So identify the user as
        # "anon" or "system".
        if not login:
            # If we have a fake request, we are at startup or in batch mode and
            # the user is "system". Else, it is "anon". At Zope startup, Appy
            # uses an Object instance as a fake request. In "zopectl run" mode
            # (the Zope batch mode), Appy adds a param "_fake_" on the request
            # object created by Zope.
            if (req.__class__.__name__ == 'Object') or \
               (hasattr(req, '_fake_') and req._fake_):
                login = 'system'
            else:
                login = 'anon'
        return login, password, ctx, place

    def getUser(self, authenticate=False, source='any'):
        '''Gets the current user. If p_authenticate is True, in addition to
           finding the logged user and returning it (=identification), we check
           if found credentials are valid (=authentication).

           If p_authenticate is True and p_source is:
           * "zodb", authentication is performed locally;
           * "ldap", authentication is performed on a LDAP (if a LDAP
                     configuration is found);
           * "sso",  authentication is performed based on HTTP headers from the
                     request, set by some SSO reverse proxy;
           * "any",  authentication is performed on the local User object, be it
                     really local or a copy of an external (ldap or sso) user.
        '''
        tool = self.appy()
        req = tool.request
        # Try first to return the user that can be cached on the request. In
        # this case, we suppose authentication has previously been done, and we
        # just return the cached user.
        if hasattr(req, 'user'): return req.user
        # Identify the user (=find its login, password and authentication
        # context). If we don't need to authenticate the user, we ask to
        # identify a user or, if impossible, a special user.
        login,password,ctx,place=self.identifyUser(alsoSpecial=not authenticate)
        # Stop here if no user was found and authentication was required
        if authenticate and not login: return
        # Now, get the User instance. If the user was identified from HTTP
        # headers (SSO), always get it via the SsoConfig.
        cfg = self.getProductConfig(True)
        if place == 'sso': source = 'sso'
        if source == 'zodb':
            # Get the User object, but only if it is a true local user
            user = tool.search1('User', noSecurity=True, login=login)
            if user and (user.source != 'zodb'): user = None # Not a local one
        elif source in ('ldap', 'sso'):
            user = None
            extSource = getattr(cfg, source)
            if extSource:
                if source == 'ldap':
                    user = extSource.getUser(tool, login, password)
                else:
                    user = extSource.getUser(tool, login,
                                             createIfNotFound=authenticate)
        elif source == 'any':
            # Get the User object, be it really local or representing an
            # external user. This way, we avoid contacting the distant source
            # every time authentication is required.
            user = tool.search1('User', noSecurity=True, login=login)
            if not user and (login != 'system'):
                self.log(USER_NOT_FOUND % login, noUser=True, type='error')
                self._disableUser(req)
                return tool.search1('User', noSecurity=True, login='anon')
        if not user: return
        login = user.login # May have been transformed by a source
        # Authenticate the user if required. In the case of SSO, authentication
        # has already been performed by the reverse proxy.
        if authenticate and (place != 'sso'):
            if (user.state == 'inactive') or \
               (not user.checkPassword(password)) or \
               (cfg.authContext and cfg.authContext.isMandatory(tool) and \
                not ctx and (place != 'basic')):
                # Disable the user
                self._disableUser(req)
                return
            # Create an authentication cookie for this user
            gutils.writeCookie(login, password, ctx, req)
        # Cache the user and some precomputed values, for performance
        req.user = user
        req.userLogin = login
        if ctx and cfg.authContext:
            cfg.authContext._setOnRequest(tool, req, ctx)
        else:
            # Initialise attribute "authContext" with the empty context
            req.authContext = None
        req.userRoles = user.getRoles()
        req.userLogins = user.getLogins()
        req.zopeUser = user.getZopeUser()
        return user

    def performLogin(self):
        '''Logs the user in'''
        rq = self.REQUEST
        jsEnabled = rq.get('js_enabled', False) in ('1', 1)
        cookiesEnabled = rq.get('cookies_enabled', False) in ('1', 1)
        if jsEnabled and not cookiesEnabled:
            msg = self.translate('enable_cookies')
            return self.goto(self.getReferer(), msg)
        # Authenticate the user
        if self.getUser(authenticate=True, source='zodb') or \
           self.getUser(authenticate=True, source='ldap'):
            user = rq.user
            suffix = user.changePasswordAtNextLogin and '_change_password' or ''
            msg = self.translate('login_ok%s' % suffix)
            logMsg = 'logged in.'
            success = True
            user.lastLoginDate = DateTime()
            # Set a default authentication context when relevant
            authContext = self.getProductConfig(True).authContext
            if authContext: authContext._setDefault(rq, self.appy())
        else:
            msg = self.translate('login_ko')
            login = rq.get('__ac_name', '').strip() or '<empty>'
            logMsg = 'authentication failed with login %s.' % login
            success = False
        self.log(logMsg)
        # Must the user be redirected to some URL ?
        gotoUrl = rq.get('goto', None)
        siteUrl = self.getSiteUrl()
        if not success:
            backUrl = siteUrl
            if gotoUrl: backUrl += '?goto=%s' % urllib.quote(gotoUrl)
        else:
            if gotoUrl:
                backUrl = '%s/%s' % (siteUrl, gotoUrl)
                msg = None # No need to get the welcome message here
            else:
                backUrl = siteUrl
                # Carry special key "authInfo" if present. It is a mechanism
                # allowing to carry arbitrary content after successful
                # authentication.
                authInfo = rq.get('authInfo')
                if authInfo: backUrl = '%s?authInfo=%s' % (backUrl, authInfo)
        return self.goto(backUrl, msg)

    def performLogout(self):
        '''Logs out the current user when he clicks on "disconnect"'''
        rq = self.REQUEST
        userId = self.getUser().login
        # Perform the logout in acl_users
        rq.RESPONSE.expireCookie('_appy_', path='/')
        # Invalidate the user session
        try:
            sdm = self.session_data_manager
        except AttributeError, ae:
            # When ran in test mode, session_data_manager is not there
            sdm = None
        if sdm:
            session = sdm.getSessionData(create=0)
            if session is not None:
                session.invalidate()
        self.log('logged out.')
        # Remove user from variable "loggedUsers"
        if self.loggedUsers.has_key(userId): del self.loggedUsers[userId]
        return self.goto(self.getApp().absolute_url())

    def getLogoutUrl(self):
        '''The "logout" URL may not be the standard Appy one if the app is
           behind a SSO reverse proxy.'''
        if self.getUser().source == 'sso':
            return self.getProductConfig(True).sso.logoutUrl
        return '%s/performLogout' % self.absolute_url()

    # This dict stores, for every logged user, the date/time of its last access
    loggedUsers = {}
    staticExtensions = ('.jpg', '.jpeg', '.gif', '.png', '.js', '.css', '.htm',
                        '.html')
    def rememberAccess(self, user):
        '''Every time there is a hit on the server, this method is called in
           order to update global dict loggedUsers (see above).'''
        now = DateTime()
        self.loggedUsers[user.login] = now
        # "Touch" the SESSION object. Else, expiration won't occur.
        try:
            session = self.REQUEST.SESSION
        except Exception, e:
            # On heavy load, for an unknown reason, an exception can be raised:
            # SessionDataManagerErr: External session data container
            # '/temp_folder/session_data' not found.
            pass
        return now

    def validate(self, request, auth='', roles=_noroles):
        '''This method performs authentication and authorization. It is used as
           a replacement for Zope's AccessControl.User.BasicUserFolder.validate,
           that allows to manage cookie-based authentication.'''
        v = request['PUBLISHED'] # The published object
        tool = self.getParentNode().config
        # v is the object (value) we're validating access to
        # n is the name used to access the object
        # a is the object the object was accessed through
        # c is the physical container of the object
        a, c, n, v = self._getobcontext(v, request)
        # Authorize anyone to static content (image, js, css...)
        id = a.getId()
        if id and (os.path.splitext(id)[-1].lower() in tool.staticExtensions):
            return self._nobody.__of__(self)
        # Skip authorization when the performing http login: else, it will be
        # done twice.
        if (id == 'config') and (v.__name__ == 'performLogin'):
            return self._nobody.__of__(self)
        # Identify and authenticate the user
        user = tool.getUser(authenticate=True, source='any')
        if not user:
            # Login and/or password incorrect. Try to authorize and return the
            # anonymous user.
            if self.authorize(self._nobody, a, c, n, v, roles):
                return self._nobody.__of__(self)
            else:
                return
        else:
            # We found a user and his password was correct. Try to authorize him
            # against the published object. By the way, remember its last access
            # to this system.
            tool.rememberAccess(user)
            user = user.getZopeUser()
            if self.authorize(user, a, c, n, v, roles):
                return user.__of__(self)
            # That didn't work. Try to authorize the anonymous user.
            elif self.authorize(self._nobody, a, c, n, v, roles):
                return self._nobody.__of__(self)
            else:
                return

    # Patch BasicUserFolder with our version of m_validate above
    from AccessControl.User import BasicUserFolder
    BasicUserFolder.validate = validate

    def getUserName(self, login=None, normalized=False):
        return self.appy().getUserName(login=login, normalized=normalized)

    def tempFile(self):
        '''A temp file has been created in a temp folder. This method returns
           this file to the browser.'''
        rq = self.REQUEST
        baseFolder = os.path.join(sutils.getOsTempFolder(), self.getAppName())
        baseFolder = os.path.join(baseFolder, rq.SESSION.id)
        fileName   = os.path.join(baseFolder, rq.get('name', ''))
        if os.path.exists(fileName):
            f = file(fileName)
            content = f.read()
            f.close()
            # Remove the temp file
            os.remove(fileName)
            return content
        return 'File does not exist'

    def getResultPodFields(self, contentType):
        '''Finds, among fields defined on p_contentType, which ones are Pod
           fields that need to be shown on a page displaying query results.'''
        return [f for f in self.getAllAppyTypes(contentType) \
                if (f.type == 'Pod') and (f.show == 'query')]

    def formatDate(self, date, format=None, withHour=True, language=None):
        '''Returns p_date formatted as specified by p_format, or tool.dateFormat
           if not specified. If p_withHour is True, hour is appended, with a
           format specified in tool.hourFormat.'''
        tool = self.appy()
        fmt = format or tool.dateFormat
        # Resolve appy-specific formatting symbols used for getting translated
        # names of days or months:
        # - %dt: translated name of day
        # - %DT: translated name of day, capitalized
        # - %mt: translated name of month
        # - %MT: translated name of month, capitalized
        # - %dd: day number, but without leading '0' if < 10
        if ('%dt' in fmt) or ('%DT' in fmt):
            day = self.translate('day_%s' % date._aday, language=language)
            fmt = fmt.replace('%dt', day.lower()).replace('%DT', day)
        if ('%mt' in fmt) or ('%MT' in fmt):
            month = self.translate('month_%s' % date._amon, language=language)
            fmt = fmt.replace('%mt', month.lower()).replace('%MT', month)
        if '%dd' in fmt: fmt = fmt.replace('%dd', str(date.day()))
        # Resolve all other, standard, symbols
        res = date.strftime(fmt)
        # Append hour from tool.hourFormat
        if withHour and (date._hour or date._minute):
            res += ' (%s)' % date.strftime(tool.hourFormat)
        return res

    def generateUid(self, className):
        '''Generates a UID for an instance of p_className.'''
        name = className.split('_')[-1]
        randomNumber = str(random.random()).split('.')[1].replace('e-', '')
        timestamp = ('%f' % time.time()).replace('.', '')
        return '%s%s%s' % (name, timestamp, randomNumber)

    def getMainPages(self):
        '''Returns the main pages.'''
        if hasattr(self.o.aq_base, 'pages') and self.o.pages:
            return [self.getObject(uid) for uid in self.o.pages ]
        return ()

    def askPasswordReinit(self):
        '''A user (anonymmous) does not remember its password. Here we will
           send him a mail containing a link that will trigger password
           re-initialisation.'''
        backUrl = self.getReferer()
        login = self.REQUEST.get('login')
        if not login: return self.goto(backUrl, self.translate('action_ko'))
        login = login.strip()
        appyTool = self.appy()
        user = appyTool.search1('User', login=login, noSecurity=True)
        msg = self.translate('reinit_mail_sent')
        if not user:
            # Return the message nevertheless. This way, malicious users can't
            # deduce information about existing users.
            return self.goto(backUrl, msg)
        # If login is an email, use it. Else, use user.email instead.
        email = user.login
        if not String.EMAIL.match(email):
            email = user.email
        if not email:
            # Impossible to re-initialise the password
            return self.goto(backUrl, msg)
        # Create a temporary file whose name is the user login and whose
        # content is a generated token.
        f = file(os.path.join(sutils.getOsTempFolder(), login), 'w')
        token = String().generatePassword()
        f.write(token)
        f.close()
        # Send an email
        initUrl = '%s/doPasswordReinit?login=%s&token=%s' % \
                  (self.absolute_url(), login, token)
        subject = self.translate('reinit_password')
        map = {'url':initUrl, 'siteUrl':self.getSiteUrl()}
        body= self.translate('reinit_password_body', mapping=map, format='text')
        appyTool.sendMail(email, subject, body)
        return self.goto(backUrl, msg)

    def doPasswordReinit(self):
        '''Performs the password re-initialisation.'''
        rq = self.REQUEST
        login = rq['login']
        token = rq['token']
        # Check if such token exists in temp folder
        res = None
        siteUrl = self.getSiteUrl()
        tokenFile = os.path.join(sutils.getOsTempFolder(), login)
        if os.path.exists(tokenFile):
            f = file(tokenFile)
            storedToken = f.read()
            f.close()
            if storedToken == token:
                # Generate a new password for this user
                appyTool = self.appy()
                user = appyTool.search1('User', login=login, noSecurity=True)
                newPassword = user.setPassword()
                # Send the new password by email
                email = login
                if not String.EMAIL.match(email):
                    email = user.email
                subject = self.translate('new_password')
                map = {'password': newPassword, 'siteUrl': siteUrl}
                body = self.translate('new_password_body', mapping=map,
                                      format='text')
                appyTool.sendMail(email, subject, body)
                os.remove(tokenFile)
                res = self.goto(siteUrl, self.translate('new_password_sent'))
        if not res:
            res = self.goto(siteUrl, self.translate('wrong_password_reinit'))
        return res

    def getGoogleAnalyticsCode(self):
        '''If the config defined a Google Analytics ID, this method returns the
           Javascript code to be included in every page, allowing Google
           Analytics to work.'''
        # Disable Google Analytics when we are in debug mode.
        if self.isDebug(): return
        # Disable Google Analytics if no ID is found in the config.
        gaId = self.getProductConfig(True).googleAnalyticsId
        if not gaId: return
        # Google Analytics must be enabled: return the chunk of Javascript
        # code specified by Google.
        code = "var _gaq = _gaq || [];\n" \
               "_gaq.push(['_setAccount', '%s']);\n" \
               "_gaq.push(['_trackPageview']);\n" \
               "(function() {\n" \
               "  var ga = document.createElement('script'); " \
               "ga.type = 'text/javascript'; ga.async = true;\n" \
               "  ga.src = ('https:' == document.location.protocol ? " \
               "'https://ssl' : 'http://www') + " \
               "'.google-analytics.com/ga.js';\n" \
               "  var s = document.getElementsByTagName('script')[0]; " \
               "s.parentNode.insertBefore(ga, s);\n" \
               "})();\n" % gaId
        return code

    def getButtonCss(self, label, small=True, render='button'):
        '''Gets the CSS class(es) to set on a button, given it l_label, its size
           (p_small or not) and its rendering (p_render).'''
        # CSS for a small button. No minimum width applies: small buttons are
        # meant to be small.
        if small:
            part = (render == 'icon') and 'Icon' or 'Small'
            return 'button%s button' % part
        # CSS for a normal button. A minimum width (via buttonFixed) is defined
        # when the label is small: it produces ranges of buttons of the same
        # width (excepted when labels are too large), which is more beautiful.
        if len(label) < 15: return 'buttonFixed button'
        return 'button'

    def getLinksTargetInfo(self, klass, back=None):
        '''Appy allows to open links to view or edit instances of p_klass
           either via the same browser window, or via a popup. Returns a
           LinkTarget instance.'''
        return gutils.LinkTarget(klass, back)

    def backFromPopup(self, initiator=None):
        '''Set the cookie allowing to close the iframe popup and refresh the
           main page.'''
        self.setMessageCookie()
        # The initiator may force to go back to some URL
        if initiator and initiator.backFromPopupUrl:
            close = base64.b64encode(initiator.backFromPopupUrl)
        else:
            close = 'yes'
        resp = self.REQUEST.RESPONSE
        resp.setCookie('closePopup', close, path='/')
        # This should not be necessary to return this Javascript call: a JS
        # timer does it, but is it not called by Edge.
        return '<html><head><script src="%s/ui/appy.js"></script></head>' \
               '<body><script>backFromPopup()</script></body></html>' % \
               self.getSiteUrl()

    ieRex = re.compile('MSIE\s+(\d\.\d)')
    ieMin = '9' # We do not support IE below this version
    def getBrowserIncompatibility(self):
        '''Produces an error message if the browser in use is not compatible
           with Appy.'''
        ua = self.REQUEST.get('HTTP_USER_AGENT')
        res = self.ieRex.search(ua)
        if not res: return
        version = res.group(1)
        if (version < self.ieMin) and ('ompatible' not in ua):
            # If term "compatible" is found, it means that we have a "modern" IE
            # but that is configured to mimick an older version. Because we have
            # set a "X-UA-Compatible" directive in all Appy pages, normally,
            # this kind of IE should not apply this old compatibility mode with
            # Appy apps, so we do not display the warning in that case.
            mapping = {'version': version, 'min': self.ieMin}
            return self.translate('wrong_browser', mapping=mapping)

    def executeAjaxAction(self, action, obj, field):
        '''When PX "pxAjax" is called to get some chunk of XHTML via an Ajax
           request, a server action can be executed before rendering the XHTML
           chunk. This method executes this action.'''
        # Find the object being the action's target. It can be p_obj or an
        # object whose ID is embedded into p_action.
        if '*' in action:
            id, action = action.split('*')
            obj = self.getObject(id, appy=True) or obj
        # Execute the action
        if action.startswith(':'):
            # The action corresponds to a method on the Appy object
            msg = getattr(obj, action[1:])()
        else:
            # The action must be executed on p_field if present, on obj.o else
            if field and hasattr(field, action):
                msg = getattr(field, action)(obj.o)
            else: msg = getattr(obj.o, action)()
        return msg

    def updatePxContextFromRequest(self):
        '''Takes any user-defined key from the request and put it as a variable
           on the current PX context.'''
        req = self.REQUEST
        ctx = req.pxContext
        # Get "form" data (get, post) and cookie values
        for source in (req.form, req.cookies):
            for k, v in source.iteritems():
                # Convert v to some Python data when relevant
                if v in ('True', 'False', 'true', 'false'):
                    exec 'v = %s' % v.capitalize()
                elif v.isdigit(): v = int(v)
                ctx[k] = v

    def patchRequestFromTemplate(self, obj, req):
        '''p_obj is a temp object that is going to be edited via pxEdit. If a
           template obj must be used to pre-fill the web form, patch the request
           with the values retrieved on this template object.'''
        if not obj.isTemporary(): return
        templateId = req.get('template')
        if not templateId: return
        template = self.getObject(templateId)
        # Some fields may be excluded from the copy
        toExclude = getattr(obj.appy().klass, 'createExclude', ())
        if callable(toExclude): toExclude = toExclude(template.appy())
        for field in template.getAllAppyTypes():
            if field.name in toExclude: continue
            field.setRequestValue(template)

    def check(self):
        '''Returns monitoring info about the running application'''
        cfg = self.getProductConfig(True)
        return cfg.monitoring.get(self.REQUEST, cfg)

    def getJsErrors(self, validator):
        '''See homonym method on appy.gen.validate.Validator'''
        r = validator and validator.getJsErrors() or 'null'
        return 'var errors = %s;' % r
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import time

from appy.px import Px
from appy.model.base import Base
from appy.model.fields import Show
from appy.ui.layout import Layouts
from appy.model.searches import Search
from appy.database.operators import or_
from appy.utils import string as sutils
from appy.model.fields.date import Date
from appy.utils.string import Normalize
from appy.model.utils import Object as O
from appy.model.fields.phase import Page
from appy.model.fields.group import Group
from appy.model.fields.string import String
from appy.model.fields.action import Action
from appy.model.fields.boolean import Boolean
from appy.model.fields.computed import Computed
from appy.model.fields.password import Password
from appy.model.workflow import standard as workflow
from appy.model.fields.select import Select, Selection

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SCAN_DB    = 'Scanning all database objects...'
LOGIN_RN   = 'User "%s" has new login "%s".'
LOGIN_OP   = "Login change: local roles and/or history updated in %d/%d " \
             "object(s). Done in %.2f''."
O_WALKED   = '%d objects walked so far...'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class User(Base):
    '''Base class representing a user'''
    workflow = workflow.Owner
    noLoginUsers = ('system', 'anon')
    specialUsers = noLoginUsers + ('admin',)
    searchAdvanced = Search('advanced', actionsDisplay='inline')
    listColumns = ('title', 'name*15%|', 'login*20%|', 'roles*20%|',
                   'state*10%|')

    # The following users are always present in any Appy database
    defaultUsers = {'admin': ('Manager',), 'system': ('Manager',), 'anon': ()}

    def isAnon(self):
        '''Is the logged user anonymous ?'''
        return self.login == 'anon'

    def isSpecial(self, includeAdmin=True):
        '''Is this user a predefined user ?'''
        attr = 'specialUsers' if includeAdmin else 'noLoginUsers'
        return self.login in getattr(User, attr)

    @staticmethod
    def update(class_):
        '''Hide the title'''
        title = class_.fields['title']
        title.show = 'xml'
        title.page.icon = 'pageProfile.svg'

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Password
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    # This page contains the single field "password" allowing to edit a local
    # user's password.

    def showPassword(self):
        '''Field "password" must be shown only for the local user allowed to
           modify it.'''
        mayEdit = (self.source == 'zodb') and (self.user.login == self.login)
        return 'edit' if mayEdit else None

    def showCancel(self):
        '''Do not show the "cancel" button on page "passwords" if the user is
           forced to change its password.'''
        return not self.changePasswordAtNextLogin

    pagePassword = Page('password', show=showPassword, showCancel=showCancel,
                        label='User_page_password', icon='pagePassword.svg')

    password = Password(multiplicity=(1,1), show=showPassword, label='User',
                        page=pagePassword)

    # The encrypted password, carriable via the XML layout
    def showEncrypted(self):
        '''Allow to show encrypted password to admins only, via the xml
           layout.'''
        if self.user.login == 'admin' and (self.source != 'sso'): return 'xml'

    encrypted = Computed(show=showEncrypted, label='User',
                         method=lambda o: o.values['password'])

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Title, name, first name
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    pm = {'page': Page('main', label='User_page_main', icon='pageProfile.svg'),
          'width': 28, 'layouts': Layouts.g, 'label': 'User',
          'group': Group('main', style='grid', hasLabel=False)}

    def getTitle(self, normalized=False, useName=True, nameFirst=False,
                 nameTransform=None, firstNameTransform=None):
        '''Returns a nice name for this user, based on available information:
           "first name"/"name" (if p_useName is True) or title or login. If
           p_normalized is True, special chars (like accents) are converted to
           ascii chars. When p_useName is True, if p_nameFirst is True, the
           result will be "name" / "first name", else it will be
           "first name" / "name".'''
        # A specified transform ("upper", "capitalize", etc) can be specified in
        # p_nameTransform and/or p_firstNameTransform.
        login = self.login
        r = None
        if useName:
            firstName = self.firstName
            name = self.name
            if firstName and name:
                # Apply a transform if requested
                if nameTransform:
                    name = eval('name.%s()' % nameTransform)
                if firstNameTransform:
                    firstName = eval('firstName.%s()' % firstNameTransform)
                # Concatenate first and last names in the right order
                r = '%s %s' % (name, firstName) if nameFirst \
                    else ('%s %s' % (firstName, name))
        r = r or self.title or login
        return Normalize.accents(r) if normalized else r

    def getTitleFromLogin(self, login, *args, **kwargs):
        '''Similar to p_getTitle, but for a user whose l_login is given (not
           p_self). If no user with this login exists, p_login is returned.'''
        user = self.search1('User', secure=False, login=login)
        return user.getTitle(*args, **kwargs) if user else login

    def updateTitle(self):
        '''Sets a title for this user'''
        self.title = self.getTitle(nameFirst=True)

    def ensureAdminIsManager(self):
        '''User "admin" must always have role "Manager"'''
        if self.id != 'admin': return
        roles = self.roles
        if 'Manager' not in roles:
            if not roles: roles = ['Manager']
            else: roles.append('Manager')
            self.roles = roles

    def showName(self):
        '''Name and first name, by default, can not be edited for non-local
           users.'''
        if (self.source != 'zodb'): return Show.E_
        return True

    name = String(show=showName, **pm)
    firstName = String(show=showName, **pm)

    def getFirstName(self):
        '''Return p_self's first name, or its login if the first name is
           undefined.'''
        return self.firstName or self.login

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Login
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    pm['multiplicity'] = (1,1)

    def showLogin(self):
        '''When must we show the login field ?'''
        if self.isTemp(): return 'edit'
        # The manager has the possibility to change the login itself (local
        # users only).
        if self.user.hasRole('Manager') and (self.source == 'zodb'):
            return True
        return Show.E_

    def validateLogin(self, login):
        '''Is this p_login valid ?'''
        # 2 cases: (1) The user is being created and has no login yet, or
        #          (2) The user is being edited and has already a login, that
        #              can potentially be changed.
        if not self.login or (login != self.login):
            # A new p_login is requested. Check if it is valid and free.
            # Some logins are not allowed.
            if login in self.specialUsers:
                return self.translate('login_reserved')
            # Check that no user or group already uses this login
            if self.count('User', login=login) or \
               self.count('Group', login=login):
                return self.translate('login_in_use')
        return True

    login = String(show=showLogin, validator=validateLogin, indexed=True, **pm)
    del pm['label']

    def getLogins(self, groupsOnly=False, compute=False, guard=None):
        '''Gets all the logins that can "match" this user: it own login
           (excepted if p_groupsOnly is True) and the logins of all the groups
           he belongs to.

           If p_compute is False, p_groupsOnly is False and p_self is the
           currently logged user, roles are retrieved from the guard, that
           caches it. Else, they are really computed.
        '''
        guard = guard or self.guard
        # Return the cached value on the guard when appropriate
        if not compute and not groupsOnly and (self.login == guard.userLogin):
            return guard.userLogins
        # Compute it
        r = [group.login for group in self.groups or ()]
        if not groupsOnly: r.append(self.login)
        return r

    def setLogin(self, old, new, reindex=True):
        '''Modifies p_self's login, from p_old to p_new'''
        start = time.time()
        self.login = new
        # Update the email if it corresponds to the login
        email = self.email
        if email == old:
            self.email = new
        # Update the title
        self.updateTitle()
        # Browse all objects of the database and update potential local roles
        # that referred to the old login.
        self.log(SCAN_DB)
        updated = total = 0
        for tree in self.H().dbConnection.root.iobjects.values():
            for o in tree.values():
                total += 1
                # Update local roles on this object
                rchanged = o.localRoles.replace(old, new)
                hchanged = o.history.replaceLogin(old, new)
                if rchanged or hchanged:
                    updated += 1
                    o.reindex()
                # Log progression
                if (total % 10000) == 0:
                    self.log(O_WALKED % total)
        self.log(LOGIN_RN % (old, new))
        self.log(LOGIN_OP % (updated, total, time.time()-start))
        # Reindex p_self, excepted if p_reindex is False
        if reindex: self.reindex()

    def getAllowedFrom(self, roles, logins):
        '''Gets for this user, the value allowing to perform searches regarding
           index "allowed". p_roles are user roles as computed by m_getRoles and
           p_logins are user logins as computed by m_getLogins.'''
        # Get the user roles. If a copy of the list is not done, user logins
        # will be added among user roles (again and again).
        r = roles[:]
        # Get the user logins
        if self.login != 'anon':
            for login in logins:
                r.append('user:%s' % login)
        return or_(*r)

    def getAllowedValue(self):
        '''Gets, for this user, the value allowing to perform searches regarding
           index "allowed".'''
        return self.getAllowedFrom(self.getRoles(), self.getLogins())

    def showEmail(self):
        '''In most cases, email is the login. Show the field only if it is not
           the case.'''
        # Is this user local ?
        isLocal = self.source == 'zodb'
        # Show the field nevertheless if it is not empty
        if isLocal and not self.isEmpty('email'): return True
        # Hide it if the login is an email
        login = self.login
        if login and String.EMAIL.match(login): return
        # Display the email (read-only if from an external source)
        if not isLocal: return Show.E_
        return True

    pm['label'] = 'User'
    email = String(show=showEmail, **pm)

    def getMailRecipient(self, normalized=True):
        '''Returns, for this user, the "recipient string" (first name, name,
           email) as can be used for sending an email.'''
        r = self.email or self.login
        # Ensure this is really an email
        if not String.EMAIL.match(r): return
        return '%s <%s>' % (self.getTitle(normalized=normalized), r)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Roles and permissions
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    pm['multiplicity'] = (0, None)
    del pm['width']

    def showRoles(self):
        '''Global roles are not visible or editable under all circumstances'''
        # Local users can't be granted global roles
        if self.container != self.tool: return
        # Only a Manager is allowed to grant global roles
        user = self.user
        if user.hasRole('Manager'): return True
        # The user itself, owner of himself, can consult his roles
        if user.hasRole('Owner', self): return Show.E_

    roles = Select(show=showRoles, indexed=True, width=40, height=10,
                   validator=Selection(lambda o: o.model.getGrantableRoles(o)),
                   render='checkbox', **pm)

    def getRoles(self, compute=False, guard=None):
        '''Returns all the global roles granted to this user. Not simply
           self.roles, but also "ungrantable roles" (like Anonymous or
           Authenticated) and roles inherited from group membership.

           If p_compute is False and p_self is the currently logged user, roles
           are retrieved from the guard, that caches it. Else, they are really
           computed.
        '''
        guard = guard or self.guard
        # Return the cached value on the guard when appropriate
        if not compute and (self.login == guard.userLogin):
            return guard.userRoles
        # Compute it
        r = list(self.roles)
        # Add ungrantable roles
        r.append('Anonymous' if self.isAnon() else 'Authenticated')
        # Add group global roles
        for group in self.groups:
            for role in group.roles:
                if role not in r: r.append(role)
        return r

    def hasRole(self, role, o=None):
        '''Has p_self this p_role? If p_o is None, check if this user has p_role
           globally; else, check if he has it in the context of p_o.

           p_role can also be a list/tuple of roles. In this case, the method
           returns True if the user has at least one of the listed roles.'''
        # Try with the user's global roles, excepted if p_o is in "local" mode
        noo = o is None
        if noo or not o.localRoles.only:
            r = sutils.stringIsAmong(role, self.getRoles())
            if noo or r: return r
        # Check now p_o(bject)'s local roles
        logins = self.getLogins()
        for login, roles in o.localRoles.items():
            if (login in logins) and sutils.stringIsAmong(role, roles):
                return True

    def ensureIsManager(self):
        '''Ensures p_self has role "Manager"'''
        roles = self.roles
        if 'Manager' not in roles:
            if not roles: roles = ['Manager']
            else: roles.append('Manager')
            self.roles = roles

    def hasPermission(self, permission, o):
        '''Has user p_self p_permission on p_o ?'''
        # What are the roles which are granted p_permission on p_o ?
        allowedRoles = o.getWorkflow().getRolesFor(o, permission)
        # Grant access based on global user roles (that include ungrantable
        # roles like Authenticated or Anonymous), excepted if p_o is in
        # "local" mode.
        if not o.localRoles.only:
            for role in self.getRoles():
                if role in allowedRoles: return True
        # Grant access based on local roles. Gets the logins of this user and
        # all its groups
        userLogins = self.getLogins()
        for login, roles in o.localRoles.items():
            # Ignore logins not corresponding to this user
            if login not in userLogins: continue
            for role in roles:
                if role in allowedRoles: return True

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Source
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    # Where is this user stored ? By default, in the ZODB. But the user can be
    # stored in an external LDAP (source='ldap').
    source = String(show='xml', default='zodb', layouts='f', label='User')

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Actions
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def doResetPassword(self, secure=True):
        '''Triggered from the UI, this method defines a new, automatically
           generated password for this user and returns it in the UI.'''
        user = self.user
        if secure and not user.allows('write'):
            self.raiseUnauthorized()
        password = self.getField('password').generate()
        self.password = password
        self.changePasswordAtNextLogin = True
        # If p_self corresponds to the currently logged user, update the
        # authentication cookie with its new password.
        if self == user:
            self.guard.Cookie.updatePassword(self.H(), password)
        return True, self.translate('new_password_text',
                                    mapping={'password': password})

    def showResetPassword(self):
        '''Action "reset password" is available to anyone having write access to
           the user, excepted the user himself.'''
        if (self.source == 'zodb') and self.guard.mayEdit(self) and \
           (self.user != self) and not self.isSpecial(includeAdmin=False):
            return 'buttons'

    resetPassword = Action(action=doResetPassword, show=showResetPassword,
                           confirm=True, label='User', icon='pagePassword.svg',
                           sicon='password.svg', render='icon')

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Hidden fields
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    hidden = {'show': False, 'layouts': 'f', 'label': 'User'}
    # For external users (source != "zodb"), we store the date of the last time
    # the external user and the local copy were synchronized.
    syncDate = Date(format=Date.WITH_HOUR, **hidden)
    # The date of the last login for this user
    lastLoginDate = Date(format=Date.WITH_HOUR, **hidden)
    # We may force a local user (source=zodb) to change its password at next
    # login (ie, for users created by an admin).
    changePasswordAtNextLogin = Boolean(**hidden)

    def getLanguage(self):
        '''Gets the language (code) for p_self'''
        handler = self.H()
        # There may be a forced language defined for everyone
        config = handler.config.ui
        if config.forcedLanguage: return config.forcedLanguage
        # Try to get the value from a cookie. Indeed, if such a cookie is
        # present, it means that the user has explicitly chosen this language
        # via the language selector.
        r = handler.req.AppyLanguage
        if r: return r
        # Try the "Accept-Language" header, which stores language preferences as
        # defined in the user's browser. Several languages can be listed, from
        # most to less wanted.
        r = handler.headers['Accept-Language']
        if not r:
            # Return the first language supported by the app
            return config.languages[0]
        # Browse prefered languages and return the first that is among app
        # language. If no language matches, return the first one as supported by
        # the app.
        supported = config.languages
        for lang in r.split(','):
            # Extract the 2-letter code
            code = None
            i = lang.find('-')
            if i != -1:
                code = lang[:i]
            else:
                i = lang.find(';')
                if i != -1:
                    code = lang[:i]
            code = (code or lang).strip()
            if code in supported:
                # Warn the user that this one has been chosen
                handler.resp.setHeader('Content-Language', code)
                return code
        # No supported language was found among user's prefered languages
        return supported[0]

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  PXs
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    # Display, in the user strip, links to the User instance of the logged user
    pxUserLink = Px('''
     <a if="cfg.userLink" href=":user.url">
      <img src=":url('user.svg')" class="icon"/>
      <span class="topText">:user.getFirstName()</span>
     </a>
     <x if="not cfg.userLink">:user.getFirstName()</x>''')

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Class methods
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    class System:
        '''Fake class representing the "system" user at places where its
           corresponding User instance is not available yet.'''
        login = 'system'
        logins = [login]
        roles = ['Manager', 'Authenticated']
        allowedValue = or_(*roles)

        def getLogins(self, **kwargs): return self.logins
        def getRoles(self, **kwargs): return self.roles
        def getAllowedFrom(self, roles, logins): return self.allowedValue
        def getLanguage(self): return 'en'
        def getInitiator(self): return
        def hasPermission(self, permission, o): return True
        def hasRole(self, role, o=None): return True

    @classmethod
    def identify(class_, guard):
        '''To identify a user means: get its login, password and authentication
           context. There are several places to look for this information: the
           Appy authentication cookie, HTTP authentication, or credentials from
           the login form.'''

        # If no user could be identified, the "anon" user, representing an
        # anonymous user, will nevertheless be identified.
        handler = guard.handler
        req = handler.req
        login = password = ctx = place = None
        # a. Identify the user from the authentication cookie
        try:
            login, password, ctx = guard.Cookie.read(handler)
        except Exception as e:
            handler.log('app', 'error', 'Unreadable cookie (%s)' % str(e))
        if login: place = 'cookie'
        # b. Identify the user from HTTP basic authentication
        if not login and ('_basic_auth_' in req):
            # Credentials from HTTP basic authentication are present, decode it
            creds = req._basic_auth_
            if creds.lower().startswith('basic '):
                try:
                    creds = creds.split(' ')[-1]
                    login, password = base64.decodestring(creds).split(':', 1)
                    if login: place = 'basic'
                except Exception:
                    pass
        # c. Identify the user from the authentication form
        if not login:
            login = req.login
            if login:
                login = handler.config.security.transformLogin(login.strip())
            password = req.password or ''
            ctx = req.context
            if login: place = 'form'
        # d. Identify the user from a SSO reverse proxy
        if not login or (login == '_s_s_o_'):
            sso = handler.config.security.sso
            if sso:
                login = sso.extractUserLogin(guard)
                if login: place = 'sso'
        # e. If all identification methods failed, identify the user as "anon"
        return login or 'anon', password, ctx, place

    @classmethod
    def authenticate(class_, guard):
        '''Authenticate the currently logged user and return its corresponding
           User instance.'''
        handler = guard.handler
        # Manage a non-HTTP request
        if handler.fake:
            # We are running in a specific context: at HTTP server startup, or
            # by a script. The "system" user, representing the running server
            # himself, must be used and returned. This user, for which a User
            # instance normally exists, may be requested before it is created
            # (on database initialization). This is why we may create here a
            # fake one.
            if hasattr(handler, 'connection'):
                user = handler.dbConnection.root.objects.get('system') or \
                       User.System()
            else:
                user = User.System()
            return user
        # Identify the user
        login, password, ctx, place = User.identify(guard)
        config = guard.config.security
        authContext = config.authContext
        tool = handler.tool
        if place == 'sso':
            # We have found a user already authenticated by a SSO (Single
            # Sign-On) reverse proxy: its credentials were carried in HTTP
            # headers. Simply return its local copy, or create it if not found.
            user = config.sso.getUser(tool, login, createIfNotFound=True)
        else:
            # In any other case, authentication must be performed
            user = None
            if place == 'form':
                # A user has typed its login and password from the ui (pxLogin).
                # If a LDAP server is defined, try to find it there.
                if config.ldap:
                    user = config.ldap.getUser(tool, login, password)
            # Get the user from the local database if it was not found yet
            user = user or tool.search1('User', login=login)
            # Authentication fails (and user "anon" is returned) if the user was
            # not found or inactive, its password was invalid or the required
            # authentication context was not found.
            if (user is None) or user.isAnon() or (user.state=='inactive') or \
               (not user.getField('password').check(user, password)) or \
               (not ctx and authContext and authContext.isMandatory(tool) and \
                (place != 'basic')):
                # Disable the authentication cookie
                guard.Cookie.disable(handler)
                user = handler.dbConnection.root.objects.get('anon')
            else:
                # The user is authenticated. Create an authentication cookie for
                # him. Set a default context when appropriate.
                if authContext and not ctx:
                    ctx = authContext.getDefaultContext(tool, user)
                guard.Cookie.write(handler, user.login, password, ctx)
        # Set, on the guard, when present, attributes related to the
        # authentication context.
        if ctx and authContext: authContext._setOnGuard(tool, user, guard, ctx)
        return user

    @classmethod
    def getOnline(class_, tool, expr=None, context=None):
        '''Returns a XHTML table showing users being currently connected'''
        # If p_expr is given, it must be an expression that will be evaluated on
        # every connected user. The user will be part ofthe result only if the
        # expression evaluates to True. The user is given to the expression as
        # variable "user".
        # ~~~
        # Get and count connected users
        users = list(tool.H().onlineUsers.items())
        users.sort(key=lambda u: u[1], reverse=True) # Sort by last access date
        count = len(users)
        # Prepare the expression's context when relevant
        if expr:
            if context is None: context = {}
        # Compute table rows
        rows = []
        for login, access in users:
            user = tool.search1('User', login=login)
            if not user: continue # Could have been deleted in the meanwhile
            if expr:
                # Update the p_expr's context with the current user
                context['user'] = user
                # Evaluate it
                if not eval(expr, context): continue
            rows.append('<tr><td><a href="%s">%s</a></td><td>%s</td></tr>' % \
                        (user.url, user.title, tool.formatDate(access)))
        # Create an empty row if no user has been found
        if not rows:
            rows.append('<tr><td colspan="2" align="center">%s</td></tr>' %
                        tool.translate('no_value'))
            count = 0
        else:
            count = len(rows)
        # Build the entire table
        r = '<table class="list"><tr><th>(%d)</th><th>%s</th></tr>' % \
            (count, tool.translate('last_user_access'))
        return r + '\n'.join(rows) + '</table>'

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    #  Appy methods
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def onEdit(self, created):
        '''To do when a user is p_created or updated'''
        # For a freshly created local user, generate him a password (if not
        # already set).
        isLocal = self.source == 'zodb'
        message = None
        if created and isLocal and self.isEmpty('password') and \
           self.config.security.generatePassword:
            # Generate a clear password
            password = self.getField('password').generate()
            # Store it, encrypted
            self.password = password
            # Return the clear password to the UI
            message = self.translate('new_password_text',
                                     mapping={'password': password})
        page = self.req.page or 'main'
        if page == 'main':
            login = self.login
            # (re)-compute p_self's title, computed from other fields
            self.updateTitle()
            # Ensure correctness of some infos about this user
            if isLocal and (self.id == 'admin'): self.ensureIsManager()
            # p_self must be owned by itself
            self.localRoles.add(login, 'Owner')
            # If the user was created by anon|system, anon|system can't stay its
            # Owner.
            self.localRoles.delete(('anon', 'system'))
        elif page == 'password':
            # If p_self corresponds to the currently logged user, update the
            # authentication cookie with its new password.
            if self.user == self:
                self.guard.Cookie.updatePassword(self.H(), self.req.password)
            # Reset this flag
            self.changePasswordAtNextLogin = False
        return message

    def mayEdit(self):
        '''No one can edit users "system" and "anon"'''
        return not self.isSpecial(includeAdmin=False)

    def mayDelete(self):
        '''Special users cannot be deleted'''
        return not self.isSpecial()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



class Student:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

students = [
   Student(parent_guardian='Parent 1', street='Street 1', city='Flawinne',
           state='Namur', zip='5020', lname='Name 1', fname='First name 1'),
   Student(parent_guardian='Parent 2', street='Street 2', city='Flawinne',
           state='Namur', zip='5020', lname='Name 2', fname='First name 2'),
   Student(parent_guardian='Parent 3', street='Street 3', city='Flawinne',
           state='Namur', zip='5020', lname='Name 3', fname='First name 3'),
   Student(parent_guardian='Parent 4', street='Street 4', city='Flawinne',
           state='Namur', zip='5020', lname='Name 4', fname='First name 4'),
   Student(parent_guardian='Parent 5', street='Street 5', city='Flawinne',
           state='Namur', zip='5020', lname='Name 5', fname='First name 5'),
]



'''Search results can be displayed in various "modes": as a list, grid,
   calendar, etc.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.model.batch import Batch
from appy.utils import string as sutils
from appy.model.utils import Object as O
from appy.ui import LinkTarget, Columns, Title
from appy.model.searches.gridder import Gridder

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Mode:
    '''Abstract base class for search modes. A concrete Mode instance is created
       every time search results must be computed.'''

    # The default mode(s) for displaying instances of any Appy class
    default = ('list',)

    # All available predefined concrete modes
    concrete = ('list', 'grid', 'calendar')

    # The list of custom actions that can be triggered on search results
    pxActions = Px('''
     <table>
      <tr><td for="action in actions"
            var2="field=action; fieldName=field.name; layout='query';
                  multi=True">:action.pxRender</td></tr>
     </table>''')

    @classmethod
    def get(class_, uiSearch):
        '''Create and return the Mode instance corresponding to the current
           result mode to apply to a list of instances.'''
        name = uiSearch.req.resultMode or uiSearch.getModes()[0]
        # Determine the concrete mode class
        if name in Mode.concrete:
            custom = False
            concrete = eval(name.capitalize())
        else:
            custom = True
            concrete = Custom
        # Create the Mode instance
        r = concrete(uiSearch)
        if custom:
            # The custom PX is named "name" on the model class
            r.px = getattr(self.class_, name)
        r.init()
        return r

    @classmethod
    def getText(class_, name, _):
        '''Gets the i18n text corresponding to mode named p_name'''
        name = name.rsplit('_', 1)[0] if '_' in name else name
        return _('result_mode_%s' % name) if name in Mode.concrete \
                                          else _('custom_%s' % name)

    def __init__(self, uiSearch):
        # The tied UI search
        self.uiSearch = uiSearch
        # The class from which we will search instances
        self.class_ = uiSearch.container
        # Are we in a popup ?
        self.popup = uiSearch.popup
        # The tool
        self.tool = uiSearch.tool
        # The ID of the tag that will be ajax-filled with search results
        self.hook = 'searchResults'
        # Matched objects
        self.objects = None
        # A Batch instance, when only a sub-set of the result set is shown at
        # once.
        self.batch = None
        # Determine result's "emptiness". If a search produces results without
        # any filter, it is considered not being empty. Consequently, a search
        # that would produce results without filters, but for which there is no
        # result, due to current filters, is not considered being empty.
        self.empty = True
        # URL for triggering a new search
        self.newSearchUrl = None
        # Is the search triggered from a Ref field ?
        self.fromRef = False
        # Is the search integrated into another field ?
        self.inField = False
        # The target for "a" tags
        self.target = LinkTarget(self.class_, popup=uiSearch.search.viaPopup)
        # How to render links to result objects ?
        self.titleMode = uiSearch.getTitleMode(self.popup)

    def init(self):
        '''Lazy mode initialisation. Can be completed by sub-classes.'''
        # Store criteria for custom searches
        tool = self.tool
        req = tool.req
        self.criteria = req.criteria
        ui = self.uiSearch
        search = ui.search
        # The search may be triggered via a Ref field
        io, ifield = search.getRefInfo(tool, nameOnly=False)
        if io:
            self.refObject = io
            self.refField = ifield
        else:
            self.refObject, self.refField = None, None
        # Build the URL allowing to trigger a new search
        if search.name == 'customSearch':
            part = '&ref=%s:%s' % (io.iid, ifield.name) if io else ''
            self.newSearchUrl = '%s/Search/advanced?className=%s%s' % \
                                (tool.url, search.container.name, part)
        self.fromRef = bool(ifield)
        self.inField = ifield and ('search*' in req.search)

    def getAjaxData(self):
        '''Initializes an AjaxData object on the DOM node corresponding to the
           ajax hook for this search result.'''
        search = self.uiSearch
        name = search.name
        params = {'className': self.class_.name, 'search': name,
                  'popup': self.popup}
        # Add initiator-specific params
        if search.initiator:
            initatorParams = search.initiator.getAjaxParams()
            if initatorParams: params.update(initatorParams)
        # Add custom search criteria
        if self.criteria:
            params['criteria'] = self.criteria
        # Add the "ref" key if present
        ref = self.tool.req.ref
        if ref:
            params['ref'] = ref
        # Concrete classes may add more parameters
        self.updateAjaxParameters(params)
        # Convert params into a JS dict
        params = sutils.getStringFrom(params)
        return "new AjaxData('%s/Search/batchResults', 'POST', %s, '%s')" % \
               (self.tool.url, params, self.hook)

    def updateAjaxParameters(self, params):
        '''To be overridden by subclasses for adding Ajax parameters
           (see m_getAjaxData above)'''

    def getAjaxDataRow(self, o, **params):
        '''Initializes an AjaxData object on the DOM node corresponding to the
           row displaying info about p_o within the results.'''
        return "new AjaxData('%s/pxResult', 'GET', %s, '%d', '%s')"% \
               (o.url, sutils.getStringFrom(params), o.iid, self.hook)

    def getRefUrl(self):
        '''When the search is triggered from a Ref field, this method returns
           the URL allowing to navigate back to the page where this Ref field
           lies.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class List(Mode):
    '''Displays search results as a table containing one row per object'''

    px = Px('''
     <table class=":class_.getResultCss(layout)" width="100%">
      <!-- Headers, with filters and sort arrows -->
      <tr if="showHeaders">
       <th for="column in mode.columns"
           var2="field=column.field" width=":column.width" align=":column.align"
           class=":(field == '_checkboxes') and mode.cbClass or ''">
        <x if="column.header">
         <img if="field == '_checkboxes'" src=":url('checkAll.svg')"
              class="clickable iconM" title=":_('check_uncheck')"
              onclick=":'toggleAllCbs(%s)' % q(mode.checkboxesId)"/>
         <x if="not column.special">
          <span class="htitle">::Px.truncateText(_(field.labelId))</span>
          <!-- Sort icons -->
          <x if="field.isSortable() and (mode.batch.total &gt; 1)">
           <img if="(mode.sortKey != field.name) or (mode.sortOrder == 'desc')"
                onclick=":'askBunchSorted(%s, %s, %s)' % \
                          (q(mode.hook), q(field.name), q('asc'))"
                src=":url('asc.svg')" class="clickable iconS"/>
           <img if="(mode.sortKey != field.name) or (mode.sortOrder == 'asc')"
                onclick=":'askBunchSorted(%s, %s, %s)' % \
                          (q(mode.hook), q(field.name), q('desc'))"
                src=":url('asc.svg')" class="clickable iconS"
                style="transform:rotate(180deg)"/>
          </x>
          <!-- Filter widget -->
          <x var="field,filterPx=field.getFilterPx()"
             if="filterPx and mode.showFilters">:filterPx</x>
          <x if="ui.Title.showSub(class_, field)">:ui.Title.pxSub</x>
         </x>
        </x>
       </th>
      </tr>

      <!-- Results -->
      <tr if="not mode.objects">
       <td colspan=":len(mode.columns)+1">::_('query_no_result')</td>
      </tr>
      <x for="o in mode.objects"
         var2="rowCss=loop.o.odd and 'even' or 'odd';
               @currentNumber=currentNumber + 1">:o.pxResult</x>
     </table>

     <!-- The button for selecting objects and closing the popup -->
     <div if="popup and mode.cbShown" align=":dleft">
      <input type="button"
             var="label=_('object_link_many'); css=ui.Button.getCss(label)"
             value=":label" class=":css"
             style=":url('linkMany.svg', bg='18px 18px')"
             onclick=":uiSearch.initiator.jsSelectMany(\
                   q, mode.sortKey, mode.sortOrder, mode.getFiltersString())"/>
     </div>

     <!-- Custom actions -->
     <x var="actions=uiSearch.search.getActions(tool)"
        if="actions and not popup">:mode.pxActions</x>

     <!-- Init checkboxes if present -->
     <script if="mode.checkboxes">:'initCbs(%s)' % q(mode.checkboxesId)</script>

     <!-- Init field focus and store object IDs in the session storage -->
     <script>:'initFocus(%s); %s;' % (q(mode.hook), mode.store)</script>''')

    def init(self):
        '''List-specific initialization'''
        Mode.init(self)
        ui = self.uiSearch
        search = ui.search
        tool = self.tool
        req = tool.req
        # Build search parameters (start number, sort and filter)
        start = int(req.start or '0')
        self.sortKey = req.sortKey
        self.sortOrder = req.sortOrder or 'asc'
        self.filters = self.class_.getFilters(tool)
        # Run the search
        self.batch = search.run(tool.H(), start=start, sortBy=self.sortKey,
          sortOrder=self.sortOrder, filters=self.filters,
          refObject=self.refObject, refField=self.refField)
        self.batch.hook = self.hook
        self.objects = self.batch.objects
        # Show sub-titles ? Show filters ?
        self.showSubTitles = eval(req.showSubTitles or 'False')
        self.showFilters = self.filters or (self.batch.total > 1)
        # Every matched object may be selected via a checkbox
        self.rootHook = ui.getRootHook()
        self.checkboxes = ui.checkboxes
        if self.fromRef and self.inField:
            # The search is a simple view of objects from a Ref field:
            # checkboxes must have IDs similar to the standard view of these
            # objects via the Ref field.
            prefix = '%s_%s' % (self.refObject.iid, self.refField.name)
        else:
            prefix = self.rootHook
        self.checkboxesId = prefix + '_objs'
        self.cbShown = ui.showCheckboxes()
        self.cbClass = '' if self.cbShown else 'hide'
        # Determine result emptiness
        self.empty = not self.objects and not self.filters
        # Compute info related to every column in the list
        self.columnLayouts = self.getColumnLayouts()
        self.columns = Columns.get(tool, self.class_, self.columnLayouts,
                                   ui.dir, addCheckboxes=self.checkboxes)
        # Get the Javascript code allowing to store IDs from batch objects in
        # the browser's session storage.
        self.store = self.batch.store(search)

    def updateAjaxParameters(self, params):
        '''List-specific ajax parameters'''
        params.update(
          {'start': self.batch.start, 'filters': self.filters,
           'sortKey': self.sortKey or '', 'sortOrder': self.sortOrder,
           'checkboxes': self.checkboxes, 'checkboxesId': self.checkboxesId,
           'total': self.batch.total,
           'resultMode': self.__class__.__name__.lower()})

    def getColumnLayouts(self):
        '''Returns the column layouts'''
        r = None
        tool = self.tool
        name = self.uiSearch.name
        # Try first to retrieve this info from a potential source Ref field
        o = self.refObject
        if o:
            field = self.refField
            r = field.getAttribute(o, 'shownInfo')
        elif ',' in name:
            id, fieldName, x = name.split(',')
            o = tool.getObject(id)
            field = o.getField(fieldName)
            if field.type == 'Ref':
                r = field.getAttribute(o, 'shownInfo')
        if r: return r
        # Try to retrieve this info via search.shownInfo
        search = self.uiSearch.search
        r = search.shownInfo
        return r if r else self.class_.getListColumns(tool)

    def getFiltersString(self):
        '''Converts dict self.filters into its string representation'''
        filters = self.filters
        if not filters: return ''
        r = []
        for k, v in filters.items():
            r.append('%s:%s' % (k, v))
        return ','.join(r)

    def inFilter(self, name, value):
        '''Returns True if this p_value, for field named p_name, is among the
           currently set p_self.filters.'''
        values = self.filters.get(name)
        if not values: return
        if isinstance(values, str):
            r = value == values
        else:
            r = value in values
        return r

    def getNavInfo(self, number):
        '''Gets the navigation string corresponding to the element having this
           p_number within the list of results.'''
        return 'search.%s.%s.%d.%d' % (self.class_.name, self.uiSearch.name, \
                                    self.batch.start + number, self.batch.total)

    def getRefUrl(self):
        '''See Mode::getRefUrl's docstring'''
        o = self.refObject
        return '%s/view?page=%s' % (o.url, self.refField.page.name)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Grid(List):
    '''Displays search results as a table containing one cell per object'''

    px = Px('''
     <!-- Filters -->
     <div class="gridFilters" if="uiSearch.search.showFilters">
      <x for="column in [c for c in mode.columns if not c.special]"
         var2="field,filterPx=column.field.getFilterPx()">
       <div if="filterPx and mode.showFilters">
        <div><label>:_(field.labelId)</label></div><x>:filterPx</x>
       </div>
      </x>
     </div>

     <!-- The grid itself -->
     <div style=":mode.gridder.getContainerStyle()">
      <div for="o in mode.objects" class="thumbnail"
           var2="mayView=guard.mayView(o)">
       <table class="thumbtable">
        <tr var="@currentNumber=currentNumber + 1" valign="top"
            for="column in mode.columns"
            var2="field=column.field; backHook='searchResults'">
         <td if="field.name == 'title'" colspan="2">:field.pxResult</td>
         <x if="(field.name != 'title') and mode.gridder.showFields">
          <td><label lfor=":field.name">::_('label', field=field)</label></td>
          <td>:field.pxResult</td>
         </x>
        </tr>
       </table>
       <div class="thumbmore">
        <img src=":url('more')" class="clickable"
             onclick="followTitleLink(this)"/>
       </div>
      </div>
      <!-- Store object IDs in the session storage -->
      <script>:mode.store</script>

      <!-- Show a message if there is no visible object, due to filters -->
      <div if="not mode.objects">::_('query_no_filtered_result')</div>
     </div>''',

     css='''.gridFilters { display: flex; flex-wrap: wrap;
                           justify-content: center; margin-top: 20px }
            .gridFilters > div { margin: 10px 20px }''',

     js='''
      followTitleLink = function(img) {
        var parent = img.parentNode.parentNode,
            atag = parent.querySelector("a[name=title]");
        atag.click();
      }
      ''')

    def __init__(self, *args):
        List.__init__(self, *args)
        # Extract the gridder defined on p_self.class_ or create a default one
        gridder = getattr(self.class_.python, 'gridder', None)
        if callable(gridder): gridder = gridder(self.tool)
        self.gridder = gridder or Gridder()

    def init(self):
        '''Lazy initialisation'''
        List.init(self)
        # Disable the possibility to show/hide sub-titles
        self.showSubTitles = True

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Calendar(Mode):
    '''Displays search results in a monthly calendar view'''

    px = Px('''<x var="layoutType='view';
                       field=tool.getField('calendar')">:field.view</x>''')

    def __init__(self, *args):
        Mode.__init__(self, *args)
        # For this calendar view to work properly, objects to show inside it
        # must have the following attributes:
        # ----------------------------------------------------------------------
        # date    | an indexed and required Date field with format =
        #         | Date.WITH_HOUR storing the object's start date and hour;
        # ----------------------------------------------------------------------
        # endDate | a not necessarily required Date field with format =
        #         | Date.WITH_HOUR storing the object's end date and hour.
        # ----------------------------------------------------------------------
        # Optionally, if objects define the following attributes, special icons
        # indicating repeated events will be shown.
        # ----------------------------------------------------------------------
        # successor   | Ref field with multiplicity = (0,1), allowing to
        #             | navigate to the next object in the list (for a repeated
        #             | event);
        # ----------------------------------------------------------------------
        # predecessor | Ref field with multiplicity = (0,1), allowing to
        #             | navigate to the previous object in the list.
        #             | "predecessor" must be the back ref of field "successor"
        #             | hereabove.
        # ----------------------------------------------------------------------

    def init(self):
        '''Creates a stub calendar field'''
        Mode.init(self)
        # Always consider the result as not empty. This way, the calendar is
        # always shown, even if no object is visible.
        self.empty = False
        # The matched objects, keyed by day. For every day, a list of entries to
        # show. Every entry is a 2-tuple (s_entryType, Object) allowing to
        # display an object at this day. s_entryType can be:
        # ----------------------------------------------------------------------
        #  "start"  | the object starts and ends at this day: display its start
        #           | hour and title;
        # ----------------------------------------------------------------------
        #  "start+" | the object starts at this day but ends at another day:
        #           | display its start hour, title and some sign indicating
        #           | that it spans on another day;
        # ----------------------------------------------------------------------
        #  "end"    | the object started at a previous day and ends at this day.
        #           | Display its title and a sign indicating this fact;
        # ----------------------------------------------------------------------
        #  "pass"   | The object started at a previous day and ends at a future
        #           | day.
        # ----------------------------------------------------------------------
        self.objects = {} # ~{s_YYYYmmdd: [(s_entryType, Object)]}~
        # Formats for keys representing dates
        self.dayKey = '%Y%m%d'
        # Format for representing hours in the UI
        self.hourFormat = '%H:%M'
        # If filters are defined from a list mode, get it
        self.filters = self.class_.getFilters(tool)

    def updateAjaxParameters(self, params):
        '''Grid-specific ajax parameters'''
        # If filters are in use, carry them
        if self.filters:
            params['filters'] = self.filters

    # For every hereabove-defined entry type, this dict stores info about how
    # to render events having this type. For every type:
    # --------------------------------------------------------------------------
    # start      | bool | Must we show the event's start hour or not ?
    # end        | bool | Must we show the event's end hour or not ?
    # css        | str  | The CSS class to add the table event tag
    # past       | bool | True if the event spanned more days in the past
    # future     | bool | True if the event spans more days in the future
    # --------------------------------------------------------------------------
    entryTypes = {
     'start':  O(start=True,  end=True,  css=None,      past=None, future=None),
     'start+': O(start=True,  end=False, css='calMany', past=None, future=True),
     'end':    O(start=False, end=True,  css='calMany', past=True, future=None),
     'pass':   O(start=False, end=False, css='calMany', past=True, future=True),
    }

    def addEntry(self, dateKey, entry):
        '''Adds an p_entry as created by m_addEntries below into self.objects
           @key p_dateKey.'''
        r = self.objects
        if dateKey not in r:
            r[dateKey] = [entry]
        else:
            r[dateKey].append(entry)

    def addEntries(self, obj):
        '''Add, in self.objects, entries corresponding to p_obj. If p_obj spans
           a single day, a single entry of the form ("start", p_obj) is added at
           the key corresponding to this day. Else, a series of entries are
           added, each of the form (s_entryType, p_obj), with the same object,
           for every day in p_obj's timespan.

           For example, for an p_obj starting at "1975/12/11 12:00" and ending
           at "1975/12/13 14:00" will produce the following entries:
              key "19751211"  >  value ("start+", obj)
              key "19751212"  >  value ("pass", obj)
              key "19751213"  >  value ("end", obj)
        '''
        # Get p_obj's start and end dates
        start = obj.date
        startKey = start.strftime(self.dayKey)
        end = obj.endDate
        endKey = end and end.strftime(self.dayKey) or None
        # Shorthand for self.objects
        r = self.objects
        if not endKey or (endKey == startKey):
            # A single entry must be added for p_obj, at the start date
            self.addEntry(startKey, ("start", obj))
        else:
            # Add one entry at the start day
            self.addEntry(startKey, ("start+", obj))
            # Add "pass" entries for every day between the start and end days
            next = start + 1
            nextKey = next.strftime(self.dayKey)
            while nextKey != endKey:
                # Add a "pass" event
                self.addEntry(nextKey, ('pass', obj))
                # Go the the next day
                next += 1
                nextKey = next.strftime(self.dayKey)
            # Add an "end" entry at the end day
            self.addEntry(endKey, ('end', obj))

    def search(self, first, grid):
        '''Performs the query, limited to the date range defined by p_grid'''
        # Performs the query, restricted to the visible date range
        last = DateTime(grid[-1][-1].strftime('%Y/%m/%d 23:59:59'))
        dateSearch = Search(date=(first, last), sortBy='date', sortOrder='asc')
        res = self.tool.executeQuery(self.className,
          search=self.uiSearch.search, maxResults='NO_LIMIT',
          search2=dateSearch, filters=self.filters)
        # Produce, in self.objects, the dict of matched objects
        for zobj in res.objects:
            self.addEntries(zobj.appy())

    def dumpObjectsAt(self, date):
        '''Returns info about the object(s) that must be shown in the cell
           corresponding to p_date.'''
        # There may be no object dump at this date
        dateStr = date.strftime(self.dayKey)
        if dateStr not in self.objects: return
        # Objects exist
        r = []
        types = self.entryTypes
        url = self.tool.getIncludeUrl
        for entryType, obj in self.objects[dateStr]:
            # Dump the event hour and title. The back hook allows to refresh the
            # whole calendar view when coming back from the popup.
            eType = types[entryType]
            # What CSS class(es) to apply ?
            css = eType.css and ('calEvt %s' % eType.css) or 'calEvt'
            # Show start and/or end hour ?
            eHour = sHour = ''
            if eType.start:
                sHour = '<td width="2em">%s</td>' % \
                        obj.date.strftime(self.hourFormat)
            if eType.end:
                endDate = obj.endDate
                if endDate:
                    eHour = ' <abbr title="%s">¬</abbr>' % \
                            endDate.strftime(self.hourFormat)
            # Display indicators that the event spans more days
            past = eType.past and '⇠ ' or ''
            future = eType.future and ' ⇢' or ''
            # The event title
            title = Title.get(o, target=self.target, popup=True,
                              backHook='configcalendar', maxChars=24)
            # Display a "repetition" icon if the object is part of a series
            hasSuccessor = obj.successor
            hasPredecessor = obj.predecessor
            if hasSuccessor or hasPredecessor:
                # For the last event of a series, show a more stressful icon
                name = not hasSuccessor and 'repeated_last' or 'repeated'
                icon = '<img src="%s" class="help" title="%s"/>' % \
                       (url(name), obj.translate(name))
            else:
                icon = ''
            # Produce the complete entry
            r.append('<table class="%s"><tr valign="top">%s<td>%s%s%s%s%s</td>'\
                     '</tr></table>' % (css,sHour,past,title,future,eHour,icon))
        return '\n'.join(r)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Custom(Mode):
    '''Displays search results via a custom PX'''

    def init(self):
        '''By default, the Custom mode performs full (unpaginated) searches'''
        r = self.tool.executeQuery(self.className, search=self.uiSearch.search,
                                   maxResults='NO_LIMIT')
        # Initialise Mode's mandatory fields
        self.objects = r.objects
        self.empty = not self.objects
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''PX stands for *P*ython *X*ML. It is a templating engine that reuses the pod
   engine to produce XML (including XHTML) from templates written as a mix of
   Python and XML.'''

# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import xml.sax
from px_parser import PxParser, PxEnvironment
from appy.pod.buffers import MemoryBuffer
from appy.shared.xml_parser import xmlPrologue, xhtmlPrologue

# Exception class --------------------------------------------------------------
class PxError(Exception): pass

# ------------------------------------------------------------------------------
class Px:
    '''Represents a (chunk of) PX code'''
    xmlPrologue = xmlPrologue
    xhtmlPrologue = xhtmlPrologue

    def __init__(self, content, isFileName=False, partial=True,
                 template=None, hook=None, prologue=None, unicode=True,
                 css=None, js=None, name=None):
        '''p_content is the PX code, as a string, or a file name if p_isFileName
           is True. If this code represents a complete XML file, p_partial is
           False. Else, we must surround p_content with a root tag to be able
           to parse it with a SAX parser.

           If this PX is based on another PX template, specify the PX template
           in p_template and the name of the p_hook where to insert this PX into
           the template PX.

           If a p_prologue is specified, it will be rendered at the start of the
           PX result.

           By default, a PX's result will be a unicode. If you want to get an
           encoded str instead, use p_unicode=False.

           You can specify, in p_css and p_js, PX-specific CSS styles and
           Javascript code, as strings. These strings will be surrounded by the
           appropriate HTML tag ("style" for p_css and "script" for p_js) and
           dumped just before the PX result. Note that if the PX is executed
           more than once, its corresponding p_css and p_js will only be dumped
           before the first PX result.
        '''
        # Get the PX content
        if isFileName:
            f = file(content)
            self.content = f.read()
            f.close()
        else:
            self.content = content
        # It this content a complete XML file, or just some part of it ?
        self.partial = partial
        # Is this PX based on a template PX ?
        self.template = template
        self.hook = hook
        # Is there some (XML, XHTML...) prologue to dump ?
        self.prologue = prologue
        # Will the result be unicode or str ?
        self.unicode = unicode
        # PX-specific CSS and JS code
        self.css = self.compact(css)
        self.js = self.compact(js)
        # A PX can have a name
        self.name = name
        # A PX can be profiled (see m_profile below)
        self.profiler = None
        # Parse the PX
        self.parse()

    def parse(self):
        '''Parses self.content and create the structure corresponding to this
           PX.'''
        if self.partial:
            # Surround the partial chunk with a root tag: it must be valid XML
            self.content = '<x>%s</x>' % self.content
        # Create a PX parser
        self.parser = PxParser(PxEnvironment(), self)
        # Parses self.content (a PX code in a string) with self.parser, to
        # produce a tree of memory buffers.
        try:
            self.parser.parse(self.content)
        except xml.sax.SAXParseException, spe:
            self.completeErrorMessage(spe)
            raise spe

    def compact(self, s):
        '''Removes single-line comments and unnecessary spaces in p_s,
           representing CSS or JS code.'''
        if not s: return
        r = []
        for line in s.split('\n'):
            line = line.strip()
            if not line.startswith('//'):
                r.append(line)
        return '\n'.join(r)

    def addCssJs(self, code, type):
        '''Adds this p_code, being CSS or JS, depending on p_type'''
        code = self.compact(code)
        existing = getattr(self, type)
        if existing:
            code = '%s\n%s' % (existing, code)
        setattr(self, type, code)

    def addCss(self, css): self.addCssJs(css, 'css')
    def addJs(self, js): self.addCssJs(js, 'js')

    def completeErrorMessage(self, parsingError):
        '''A p_parsingError occurred. Complete the error message with the
           erroneous line from self.content.'''
        # Split lines from self.content
        splitted = self.content.split('\n')
        i = parsingError.getLineNumber() - 1
        # Get the erroneous line, and add a subsequent line for indicating
        # the erroneous column.
        column = ' ' * (parsingError.getColumnNumber()-1) + '^'
        lines = [splitted[i], column]
        # Get the previous and next lines when present.
        if i > 0: lines.insert(0, splitted[i-1])
        if i < len(splitted)-1: lines.append(splitted[i+1])
        parsingError._msg += '\n%s' % '\n'.join(lines)

    def __call__(self, context, applyTemplate=True, isTemplate=False):
        '''Renders the PX.

           If the PX is based on a template PX, we have 2 possibilities.
           1. p_applyTemplate is True. This case corresponds to the initial
              call to the current PX. In this case we call the template with a
              context containing, in the hook variable, the current PX.
           2. p_applyTemplate is False. In this case, we are currently executing
              the PX template, and, at the hook, we must include the current PX,
              as is, without re-applying the template (else, an infinite
              recursion would occur).
        '''
        # Developer, forget the following line
        if '_ctx_' not in context: context['_ctx_'] = context
        # Include, in the context, variable names for "reserved" PX chars
        if 'PIPE' not in context:
            context['PIPE'] = '|'
            context['SEMICOLON'] = ';'
        # Also include the PX itself and remember the previously running PX
        px = context.get('_px_')
        if not isTemplate:
            context['_px_'] = self

        # This PX call is probably one among a series of such calls, sharing the
        # same p_context. Within this context, we add a sub-dict at key "rt"
        # (*r*un-*t*ime) for counting the number of times every PX is called.
        # It allows us to include PX-specific CSS and JS code only once.
        if '_rt_' not in context: context['_rt_'] = {}

        if self.hook and applyTemplate:
            # Call the template PX, filling the hook with the current PX
            context[self.hook] = self
            r = self.template(context, isTemplate=True)
        else:
            # Start profiling when relevant
            profiler = self.profiler
            if profiler: profiler.enter(self.name)
            # Create a Memory buffer for storing the result
            env = self.parser.env
            result = MemoryBuffer(env, None)
            # Execute the PX
            env.ast.evaluate(result, context)
            # Get the PX result
            r = result.content
            # Count this call and include CSS and JS code when relevant
            pxId = id(self)
            rt = context['_rt_']
            if pxId in rt:
                rt[pxId] += 1
            else:
                rt[pxId] = 1
                # This is the first time we execute it: include CSS and JS code
                # if present.
                if self.js: r = ('<script>%s</script>\n' % self.js) + r
                if self.css: r = ('<style>%s</style>\n' % self.css) + r
            # Include the prologue and manage encoding
            if self.prologue:
                r = self.prologue + r
            if not self.unicode:
                r = r.encode('utf-8')
            # Stop profiling when relevant
            if profiler: profiler.leave()
        # Restore the previous PX in the context
        if px: context['_px_'] = px
        # Return the result
        return r

    def override(self, content, partial=True):
        '''Overrides the content of this PX with a new p_content (as a
           string).'''
        self.partial = partial
        self.content = content
        # Parse again, with new content
        self.parse()

    def profile(self, name, profiler):
        '''Enables profiling of this PX, that will be named p_name in the
           p_profiler's output.'''
        self.name = name
        self.profiler = profiler

    @staticmethod
    def setBaseVariables(tool, context, obj):
        '''Sets, in p_context, all base variables required to execute standard
           Appy PXs.'''
        # The base object, available everywhere, is the tool
        ztool = tool.o
        # The request object (may be fake)
        req = ztool.REQUEST
        # Are we in the Appy popup ?
        inPopup = req.get('popup') == '1'
        # The currently published object and logged user
        obj = obj or ztool.getHomeObject(inPopup)
        zobj = obj and obj.o or None
        user = tool.user
        # The user language (2-letters ISO code)
        lang = ztool.getUserLanguage()
        # Is this language LTR or RTL ?
        dir = ztool.getLanguageDirection(lang)
        # The root database application object
        app = ztool.getApp()
        # Define the context
        r = {'tool':tool, 'ztool':tool.o, 'req':req, 'resp':req.RESPONSE,
             'obj':obj, 'zobj':zobj, 'inPopup':inPopup, '_':ztool.translate,
             'user':user, 'isAnon':user.login == 'anon', 'q':ztool.quote,
             'app':app, 'appName':ztool.getAppName(), 'appFolder':app.data,
             'url':ztool.getIncludeUrl, 'cfg':ztool.getProductConfig(True),
             'lang':lang, 'dir':dir,
             'dleft':(dir == 'ltr') and 'left' or 'right',
             'dright':(dir == 'ltr') and 'right' or 'left'}
        # Update p_context with the created context
        context.update(r)
        # Set the PX context available on the request object, too
        req.pxContext = context

    # Names of the base variables that must be present in the context of any
    # standard Appy PX.
    contextKeys = ('tool', 'ztool', 'req', 'resp', 'inPopup', '_', 'user',
                   'isAnon', 'q', 'app', 'appName', 'appFolder', 'url', 'cfg',
                   'lang', 'dir', 'dleft', 'dright')

    @classmethod
    def copyBaseVariables(klass, context, other):
        '''Copy base variables whose names are defined in p_klass.contextKeys
           from p_other to p_context.'''
        for key in klass.contextKeys:
            context[key] = other[key]
# ------------------------------------------------------------------------------



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import os, re, httplib, sys, stat, urlparse, time, socket, xml.sax, md5, random
import urllib
from StringIO import StringIO
from mimetypes import guess_type
from base64 import encodestring
from appy import Object
from appy.shared.utils import copyData, sequenceTypes, normalizeString
from appy.shared.xml_parser import XmlUnmarshaller, XmlMarshaller

# ------------------------------------------------------------------------------
class ResourceError(Exception): pass

# ------------------------------------------------------------------------------
class FormDataEncoder:
    '''Allows to encode form data for sending it through a HTTP request'''
    def __init__(self, data):
        self.data = data # The data to encode, as a dict

    def marshalValue(self, name, value):
        if isinstance(value, basestring):
            return '%s=%s' % (name, urllib.quote(str(value)))
        elif isinstance(value, float):
            return '%s:float=%s' % (name, value)
        elif isinstance(value, int):
            return '%s:int=%s' % (name, value)
        elif isinstance(value, long):
            res = '%s:long=%s' % (name, value)
            if res[-1] == 'L':
                res = res[:-1]
            return res
        else:
            raise Exception('Cannot encode value %s' % str(value))

    def encode(self):
        res = []
        for name, value in self.data.iteritems():
            res.append(self.marshalValue(name, value))
        return '&'.join(res)

# ------------------------------------------------------------------------------
class SoapDataEncoder:
    '''Allows to encode SOAP data for sending it through a HTTP request.'''
    namespaces = {'s'  : 'http://schemas.xmlsoap.org/soap/envelope/',
                  'xsd': 'http://www.w3.org/2001/XMLSchema',
                  'xsi': 'http://www.w3.org/2001/XMLSchema-instance'}
    namespacedTags = {'Envelope': 's', 'Body': 's', '*': 'py'}

    def __init__(self, data, namespace='http://appyframework.org'):
        self.data = data
        # p_data can be:
        # - a string already containing a complete SOAP message
        # - a Python object, that we will convert to a SOAP message
        # Define the namespaces for this request
        self.ns = self.namespaces.copy()
        self.ns['py'] = namespace

    def encode(self):
        # Do nothing if we have a SOAP message already
        if isinstance(self.data, basestring): return self.data
        # self.data is here a Python object. Wrap it in a SOAP Body.
        soap = Object(Body=self.data)
        # Marshall it
        marshaller = XmlMarshaller(rootTag='Envelope', namespaces=self.ns,
                                   namespacedTags=self.namespacedTags)
        return marshaller.marshall(soap)

# ------------------------------------------------------------------------------
class JsonDecoder:
    '''Converts JSON data into Pyton data structures'''
    # Boolean and None values are different in JSON and Python
    context = {'true': True, 'false': False, 'null': None}
    UNREADABLE = 'Unreadable JSON string: %s'

    @classmethod
    def convertDict(klass, d):
        '''Returns a appy.Object instance representing dict p_d'''
        res = Object()
        for name, value in d.iteritems():
            # Ensure "name" will be a valid attribute name for a Python object
            n = normalizeString(name, 'alphanum_')
            setattr(res, n, klass.convertValue(value))
        return res

    @classmethod
    def convertList(klass, l):
        '''Every item being a dict in p_l is converted to an object'''
        i = len(l) - 1
        while i >= 0:
            l[i] = klass.convertValue(l[i])
            i -= 1

    @classmethod
    def convertValue(klass, val):
        '''Converts a JSON p_val into a Python value'''
        if isinstance(val, list):
            klass.convertList(val)
            res = val
        elif isinstance(val, dict):
            res = klass.convertDict(val)
        else:
            # In all other cases, no conversion is needed
            res = val
        return res

    @classmethod
    def decode(klass, jsonData):
        '''Converts JSON data received in a string (p_jsonData) to a Python data
           structure. JSON dicts are converted to Python objects.'''
        try:
            return klass.convertValue(eval(jsonData, klass.context))
        except SyntaxError, se:
            # The presence of char "\r" may pose problem
            jsonData = jsonData.replace('\r', '')
            try:
                return klass.convertValue(eval(jsonData, klass.context))
            except SyntaxError, err:
                raise ResourceError(JsonDecoder.UNREADABLE % jsonData)

# ------------------------------------------------------------------------------
class DigestRealm:
    '''Represents information delivered by a server requiring Digest-based
       HTTP authentication.'''
    rex = re.compile('(\w+)="(.*?)"')
    # List of attributes whose values must not be quoted
    unquoted = ('algorithm', 'qop', 'nc')

    def md5(self, value):
        '''Produces the MD5 message digest for p_value, as a hexadecimal
           32-chars string.'''
        # Will use hashlib
        return md5.new(str(value)).hexdigest()

    def __init__(self, info):
        '''p_info is the content of header key WWW-Authenticate. This
           constructor parses it and unwraps data into self's attributes.'''
        for name, value in self.rex.findall(info):
            # For attribute "qpop", split values into a list
            if name == 'qop':
                value = value.split(',')
                for i in range(len(value)):
                    value[i] = value[i].strip()
            setattr(self, name, value)
        # Set some default values
        if not hasattr(self, 'algorithm'):
            self.algorithm = 'MD5'

    def buildCredentials(self, resource, uri, httpMethod='GET'):
        '''Builds credentials to transmit to the server'''
        login = resource.username
        realm = self.realm
        algorithm = self.algorithm
        nonce = self.nonce
        # Get the "path" part of the URI (compatible with various Python)
        parsed = urlparse.urlparse(uri)
        uri = isinstance(parsed, tuple) and parsed[2] or parsed.path
        # Compute a client random nouce
        cnonce = self.md5(random.random())
        # Collect credentials info in a dict
        res = {'username': login, 'uri': uri, 'realm': realm, 'nonce': nonce,
               'algorithm': algorithm}
        # Add optional attribute "opaque"
        if hasattr(self, 'opaque'): res['opaque'] = self.opaque
        # Precompute the "HA1" part of the response, that depends on the
        # algorithm in use (MD5 or MD5-sess).
        ha1 = self.md5('%s:%s:%s' % (login, realm, resource.password))
        if algorithm == 'MD5-sess':
            ha1 = self.md5('%s:%s:%s' % (ha1, nonce, cnonce))
        # Take into account the quality of protection (qop)
        hasQop = hasattr(self, 'qop')
        if hasQop:
            qop = res['qop'] = self.qop[0]
            res['cnonce'] = cnonce
            res['nc'] = '00000001'
        else:
            qop = 'auth'
        # Precompute the "HA2" part of the response, that depends on qop
        if qop == 'auth-int':
            entity = self.md5('entity')
            ha2 = self.md5('%s:%s:%s' % (httpMethod, uri, entity))
        else:
            ha2 = self.md5('%s:%s' % (httpMethod, uri))
        # Compute the complete response
        if hasQop:
            response = self.md5('%s:%s:%s:%s:%s:%s' % \
                                (ha1, nonce, res['nc'], cnonce, qop, ha2))
        else:
            response = self.md5('%s:%s:%s' % (ha1, nonce, ha2))
        res['response'] = response
        # Convert the dict to a formatted list, quoting values when relevant
        attrs = []
        for name, value in res.iteritems():
            if name not in self.unquoted:
                value = '"%s"' % value
            attrs.append('%s=%s' % (name, value))
        # Produce the final value
        return 'Digest %s' % ', '.join(attrs)

    def __repr__(self):
        pairs = ['%s=%s' % (k, v) for k, v in self.__dict__.iteritems()]
        return '<Realm %s>' % ','.join(pairs)

# ------------------------------------------------------------------------------
class HttpResponse:
    '''Stores information about a HTTP response'''
    def __init__(self, response, body, duration=None, utf8=True,
                 responseType=None):
        self.code = response.status # The return code, ie 404, 200, 500...
        self.text = response.reason # Textual description of the code
        self.headers = response.msg # A dict-like object containing the headers
        self.body = body # The body of the HTTP response
        # p_duration, if given, is the time, in seconds, we have waited, before
        # getting this response after having sent the request.
        self.duration = duration
        self.utf8 = utf8
        # The following attribute may contain specific data extracted from
        # the previous fields. For example, when response if 302 (Redirect),
        # self.data contains the URI where we must redirect the user to.
        self.data = self.extractData(responseType)
        self.response = response

    def __repr__(self, complete=False):
        duration = suffix = ''
        if self.duration: duration = ', got in %.4f seconds' % self.duration
        if complete:
            suffix = '\n*Headers* %s\n*Body* %s\n' %  \
                     (str(self.headers), str(self.body))
        return '<HttpResponse %s (%s)%s%s>' % \
               (self.code, self.text, duration, suffix)
    def get(self): return self.__repr__(complete=True)

    def extractContentType(self, contentType):
        '''Extract the content type from the HTTP header, potentially removing
           encoding-related data.'''
        i = contentType.find(';')
        if i != -1: return contentType[:i]
        return contentType

    xmlHeaders = ('text/xml', 'application/xml', 'application/soap+xml')
    def extractData(self, responseType=None):
        '''This method extracts, from the various parts of the HTTP response,
           some useful information:
           * it will find the URI where to redirect the user to if self.code
             is 302;
           * it will return authentication-related data, if present, if
             self.code is 401;
           * it will unmarshall XML or JSON data into Python objects;
           * ...'''
        # Extract information from HTTP headers when relevant
        headers = self.headers
        if self.code == 302:
            return headers['location']
        elif self.code == 401:
            authInfo = headers.get('WWW-Authenticate')
            return authInfo and DigestRealm(authInfo) or None
        # Determine the response type from the HTTP response, or, if not found,
        # use p_responseType that may have been given.
        if headers.has_key('content-type'):
            responseType = headers['content-type']
        if not responseType: return
        # Apply some transform on the response content depending on its type
        contentType = self.extractContentType(responseType)
        # Manage JSON content
        if contentType == 'application/json':
            return JsonDecoder.decode(self.body)
        # Manage XML content
        for xmlHeader in self.xmlHeaders:
            # Ensure this is XML
            if not contentType.startswith(xmlHeader): continue
            # Return an unmarshalled version of the XML content, for easy use
            try:
                parser = XmlUnmarshaller(utf8=self.utf8)
                res = parser.parse(self.body)
                if parser.rootTag == 'exception':
                    # This is an exception: "res" contains the traceback
                    raise ResourceError('Distant server exception: %s' % res)
                return res
            except xml.sax.SAXParseException, se:
                raise ResourceError('Invalid XML response (%s)' % str(se))

# ------------------------------------------------------------------------------
urlRex = re.compile(r'(http[s]?)://([^:/]+)(:[0-9]+)?(/.+)?', re.I)
binaryRex = re.compile(r'[\000-\006\177-\277]')

class Resource:
    '''Every instance of this class represents some web resource accessible
       through HTTP.'''
    Error = ResourceError
    TIMEOUT_ERROR = 'Timed out after %d second(s).'
    standardPorts = {'http': 80, 'https': 443}

    def __init__(self, url, username=None, password=None, measure=False,
                 utf8=True, authMethod='Basic', timeout=10):
        self.username = username
        self.password = password
        self.authMethod = authMethod
        self.url = url
        # A timeout (in seconds) used when sending blocking requests to the
        # resource.
        self.timeout = timeout
        # If p_measure is True, we will measure, for every request sent, the
        # time we wait until we receive the response.
        self.measure = measure
        # If measure is True, we will store hereafter, the total time (in
        # seconds) spent waiting for the server for all requests sent through
        # this resource object.
        self.serverTime = 0
        # Split the URL into its components
        res = urlRex.match(url)
        if res:
            protocol, host, port, uri = res.groups()
            self.protocol = protocol
            self.host = host
            if port:
                self.port = int(port[1:])
            else:
                self.port = Resource.standardPorts[self.protocol]
            self.uri = uri or '/'
        else: raise ResourceError('Wrong URL: %s' % str(url))
        # If some headers must be sent with any request sent through this
        # resource (like a cookie), you can store them in the following dict.
        self.headers = {'Host': self.getHeaderHost(), 'User-Agent': 'Appy'}
        self.utf8 = utf8

    def getHeaderHost(self):
        '''Gets the content of header key "Host"'''
        # Insert the port number if not standard
        if self.port != Resource.standardPorts[self.protocol]:
            suffix = ':%d' % self.port
        else:
            suffix = ''
        return '%s%s' % (self.host, suffix)

    def __repr__(self):
        return '<Resource at %s>' % self.url

    def completeHeaders(self, headers):
        # Get standard header values from self.headers if not defined in
        # p_headers
        if self.headers:
            for k, v in self.headers.iteritems():
                if k not in headers:
                    headers[k] = v
        # Add credentials-related headers when relevant
        if not (self.username and self.password): return
        if headers.has_key('Authorization'): return
        if self.authMethod == 'Basic':
            credentials = '%s:%s' % (self.username, self.password)
            credentials = credentials.replace('\012', '')
            headers['Authorization'] = "%s %s" % (self.authMethod,
                                                  encodestring(credentials))
        headers['Connection'] = 'close'
        headers['Accept'] = '*/*'

    def send(self, method, uri, body=None, headers=None, bodyType=None,
             responseType=None):
        '''Sends a HTTP request with p_method, for p_uri.'''
        connectionClass = (self.protocol == 'https') and \
                          httplib.HTTPSConnection or httplib.HTTPConnection
        conn = connectionClass(self.host, self.port)
        # Python2.4 > Python2.7 compliance
        conn.timeout = self.timeout
        try:
            conn.connect()
        except socket.gaierror, sge:
            raise ResourceError('Check your Internet connection (%s)'% str(sge))
        except socket.timeout, te:
            raise ResourceError(self.TIMEOUT_ERROR % conn.timeout)
        except socket.error, se:
            raise ResourceError('Connection error (%s)' % str(se))
        # Tell what kind of HTTP request it will be
        conn.putrequest(method, uri, skip_host=True)
        # Add HTTP headers
        if headers == None: headers = {}
        self.completeHeaders(headers)
        for k, v in headers.iteritems(): conn.putheader(k, v)
        conn.endheaders()
        # Add HTTP body
        if body:
            if not bodyType: bodyType = 'string'
            copyData(body, conn, 'send', type=bodyType)
        # Send the request, get the reply
        if self.measure: startTime = time.time()
        try:
            response = conn.getresponse()
        except socket.timeout, te:
            raise ResourceError(self.TIMEOUT_ERROR % conn.timeout)
        if self.measure: endTime = time.time()
        body = response.read()
        conn.close()
        # Return a smart object containing the various parts of the response
        duration = None
        if self.measure:
            duration = endTime - startTime
            self.serverTime += duration
        return HttpResponse(response, body, duration=duration, utf8=self.utf8,
                            responseType=responseType)

    def mkdir(self, name):
        '''Creates a folder named p_name in this resource.'''
        folderUri = self.uri + '/' + name
        #body = '<d:propertyupdate xmlns:d="DAV:"><d:set><d:prop>' \
        #       '<d:displayname>%s</d:displayname></d:prop></d:set>' \
        #       '</d:propertyupdate>' % name
        return self.send('MKCOL', folderUri)

    def delete(self, name):
        '''Deletes a file or a folder (and all contained files if any) named
           p_name within this resource.'''
        toDeleteUri = self.uri + '/' + name
        return self.send('DELETE', toDeleteUri)

    def add(self, content, type='fileName', name=''):
        '''Adds a file in this resource. p_type can be:
           - "fileName"  In this case, p_content is the path to a file on disk
                         and p_name is ignored;
           - "zope"      In this case, p_content is an instance of
                         OFS.Image.File and the name of the file is given in
                         p_name.
        '''
        if type == 'fileName':
            # p_content is the name of a file on disk
            size = os.stat(content)[stat.ST_SIZE]
            body = file(content, 'rb')
            name = os.path.basename(content)
            fileType, encoding = guess_type(content)
            bodyType = 'file'
        elif type == 'zope':
            # p_content is a "Zope" file, ie a OFS.Image.File instance
            # p_name is given
            fileType = content.content_type
            encoding = None
            size = content.size
            body = content
            bodyType = 'zope'
        fileUri = self.uri + '/' + name
        headers = {'Content-Length': str(size)}
        if fileType: headers['Content-Type'] = fileType
        if encoding: headers['Content-Encoding'] = encoding
        res = self.send('PUT', fileUri, body, headers, bodyType=bodyType)
        # Close the file when relevant
        if type =='fileName': body.close()
        return res

    def get(self, uri=None, headers=None, params=None, followRedirect=True,
            responseType=None):
        '''Perform a HTTP GET on the server. Parameters can be given as a dict
           in p_params. p_responseType will be used if no "content-type" key is
           found on the HTTP response.'''
        if not uri: uri = self.uri
        # Encode and append params if given
        if params:
            sep = ('?' in uri) and '&' or '?'
            uri = '%s%s%s' % (uri, sep, urllib.urlencode(params))
        res = self.send('GET', uri, headers=headers, responseType=responseType)
        # Follow redirect when relevant
        if (res.code == 302) and followRedirect:
            return self.get(res.data, headers=headers)
        # Perform Digest-based authentication when relevant
        if (res.code == 401) and (self.authMethod == 'Digest') and res.data:
            # Re-trigger the request with the correct authentication headers
            headers = headers or {}
            headers['Authorization'] = res.data.buildCredentials(self, uri)
            return self.get(uri=uri, headers=headers, params=params,
                       followRedirect=followRedirect, responseType=responseType)
        return res
    rss = get

    def post(self, data=None, uri=None, headers=None, encode='form'):
        '''Perform a HTTP POST on the server. If p_encode is "form", p_data is
           considered to be a dict representing form data that will be
           form-encoded. Else, p_data will be considered as the ready-to-send
           body of the HTTP request.'''
        if not uri: uri = self.uri
        if headers == None: headers = {}
        # Prepare the data to send
        if encode == 'form':
            # Format the form data and prepare headers
            body = FormDataEncoder(data).encode()
            headers['Content-Type'] = 'application/x-www-form-urlencoded'
        else:
            body = data
        headers['Content-Length'] = str(len(body))
        return self.send('POST', uri, headers=headers, body=body)

    def soap(self, data, uri=None, headers=None, namespace=None,
             soapAction=None):
        '''Sends a SOAP message to this resource. p_namespace is the URL of the
           server-specific namespace. If header value "SOAPAction" is different
           from self.url, specify it in p_soapAction.'''
        if not uri: uri = self.uri
        # Prepare the data to send
        data = SoapDataEncoder(data, namespace).encode()
        if headers == None: headers = {}
        headers['SOAPAction'] = soapAction or self.url
        # Content-type could be 'text/xml'
        headers['Content-Type'] = 'application/soap+xml;charset=UTF-8'
        res = self.post(data, uri, headers=headers, encode=None)
        # Unwrap content from the SOAP envelope
        if hasattr(res.data, 'Body'):
            res.data = res.data.Body
        return res
# ------------------------------------------------------------------------------
